//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Core/Computation/DWBAComputation.h
//! @brief     Defines class DWBAComputation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_CORE_COMPUTATION_DWBACOMPUTATION_H
#define BORNAGAIN_CORE_COMPUTATION_DWBACOMPUTATION_H

#include "Core/Computation/IComputation.h"
#include <vector>

class DiffuseElement;
class GISASSpecularContribution;
class ParticleLayoutContribution;
class reSample;
class RoughMultiLayerContribution;
class SimulationOptions;

//! Provides function that performs a single-threaded DWBA computation
//! for a range of DiffuseElement%s.
//!
//! Handles progress counter and interrupts, runs loop over DiffuseElement%s,
//! delegates all further work to DWBATerm.
//!
//! Controlled by the multi-threading machinery in ISimulation::runSingleSimulation().

class DWBAComputation : public IComputation {
public:
    DWBAComputation(const reSample& re_sample, const SimulationOptions& options,
                    ProgressHandler& progress,
                    std::vector<std::unique_ptr<DiffuseElement>>::iterator begin_it,
                    std::vector<std::unique_ptr<DiffuseElement>>::iterator end_it);
    ~DWBAComputation() override;

private:
    //! Performs a single-threaded DWBA computation for a DiffuseElement%s
    //! in range m_begin_it ... m_end_it. Results are stored in those elements.
    void runProtected() override;

    std::vector<std::unique_ptr<DiffuseElement>>::iterator m_begin_it, m_end_it;
    const std::unique_ptr<const GISASSpecularContribution> m_specular_contrib;
    const std::unique_ptr<const RoughMultiLayerContribution> m_roughness_contrib;
    const std::vector<std::unique_ptr<const ParticleLayoutContribution>> m_layout_contribs;
};

#endif // BORNAGAIN_CORE_COMPUTATION_DWBACOMPUTATION_H
#endif // USER_API
