//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/CatSample/FormFactorItemCatalog.cpp
//! @brief     Implements class FormFactorItemCatalog
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/CatSample/FormFactorItemCatalog.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Sample/FormFactorItems.h"

FormFactorItem* FormFactorItemCatalog::create(Type type)
{
#define CREATE(t)                                                                                  \
    case Type::t:                                                                                  \
        return new t##Item()

    switch (type) {
        CREATE(Pyramid2);
        CREATE(BarGauss);
        CREATE(BarLorentz);
        CREATE(Box);
        CREATE(Cone);
        CREATE(Pyramid6);
        CREATE(Bipyramid4);
        CREATE(Cylinder);
        CREATE(Dodecahedron);
        CREATE(EllipsoidalCylinder);
        CREATE(FullSphere);
        CREATE(FullSpheroid);
        CREATE(HemiEllipsoid);
        CREATE(Icosahedron);
        CREATE(Prism3);
        CREATE(Prism6);
        CREATE(Pyramid4);
        CREATE(CosineRippleBox);
        CREATE(CosineRippleGauss);
        CREATE(CosineRippleLorentz);
        CREATE(SawtoothRippleBox);
        CREATE(SawtoothRippleGauss);
        CREATE(SawtoothRippleLorentz);
        CREATE(Pyramid3);
        CREATE(TruncatedCube);
        CREATE(TruncatedSphere);
        CREATE(TruncatedSpheroid);
    default:
        ASSERT(false);
    }

#undef CREATE
}

QVector<FormFactorItemCatalog::Type> FormFactorItemCatalog::types()
{
    return {Type::Pyramid2,
            Type::BarGauss,
            Type::BarLorentz,
            Type::Box,
            Type::Cone,
            Type::Pyramid6,
            Type::Bipyramid4,
            Type::Cylinder,
            Type::Dodecahedron,
            Type::EllipsoidalCylinder,
            Type::FullSphere,
            Type::FullSpheroid,
            Type::HemiEllipsoid,
            Type::Icosahedron,
            Type::Prism3,
            Type::Prism6,
            Type::Pyramid4,
            Type::CosineRippleBox,
            Type::CosineRippleGauss,
            Type::CosineRippleLorentz,
            Type::SawtoothRippleBox,
            Type::SawtoothRippleGauss,
            Type::SawtoothRippleLorentz,
            Type::Pyramid3,
            Type::TruncatedCube,
            Type::TruncatedSphere,
            Type::TruncatedSpheroid};
}

QVector<FormFactorItemCatalog::Type> FormFactorItemCatalog::hardParticleTypes()
{
    return {Type::Pyramid2,
            Type::Box,
            Type::Cone,
            Type::Pyramid6,
            Type::Bipyramid4,
            Type::Cylinder,
            Type::Dodecahedron,
            Type::EllipsoidalCylinder,
            Type::FullSphere,
            Type::FullSpheroid,
            Type::HemiEllipsoid,
            Type::Icosahedron,
            Type::Prism3,
            Type::Prism6,
            Type::Pyramid4,
            Type::Pyramid3,
            Type::TruncatedCube,
            Type::TruncatedSphere,
            Type::TruncatedSpheroid};
}


QVector<FormFactorItemCatalog::Type> FormFactorItemCatalog::rippleTypes()
{
    return {Type::Box,
            Type::BarGauss,
            Type::BarLorentz,
            Type::CosineRippleBox,
            Type::CosineRippleGauss,
            Type::CosineRippleLorentz,
            Type::SawtoothRippleBox,
            Type::SawtoothRippleGauss,
            Type::SawtoothRippleLorentz};
}

FormFactorItemCatalog::UiInfo FormFactorItemCatalog::uiInfo(Type type)
{
    switch (type) {
    case Type::Pyramid2:
        return {"Anisotropic pyramid", ":/SampleDesignerToolbox/images/ff_Pyramid2_64x64.png",
                "Truncated pyramid with a rectangular base"};

    case Type::BarGauss:
        return {"BarGauss", ":/SampleDesignerToolbox/images/ff_Box_64x64.png",
                "Rectangular cuboid"};

    case Type::BarLorentz:
        return {"BarLorentz", ":/SampleDesignerToolbox/images/ff_Box_64x64.png",
                "Rectangular cuboid"};

    case Type::Box:
        return {"Box", ":/SampleDesignerToolbox/images/ff_Box_64x64.png", "Rectangular cuboid"};

    case Type::Cone:
        return {"Cone", ":/SampleDesignerToolbox/images/ff_Cone_64x64.png",
                "Truncated cone with circular base"};

    case Type::Pyramid6:
        return {"Pyramid6", ":/SampleDesignerToolbox/images/ff_Pyramid6_64x64.png",
                "A truncated pyramid, based on a regular hexagon"};

    case Type::Bipyramid4:
        return {"Bipyramid4", ":/SampleDesignerToolbox/images/ff_Bipyramid4_64x64.png",
                "Compound of two truncated pyramids with a common square base "
                "and opposite orientations"};

    case Type::Dodecahedron:
        return {"Dodecahedron", ":/SampleDesignerToolbox/images/ff_Dodecahedron_64x64.png",
                "Dodecahedron"};

    case Type::Cylinder:
        return {"Cylinder", ":/SampleDesignerToolbox/images/ff_Cylinder_64x64.png",
                "Cylinder with a circular base"};

    case Type::EllipsoidalCylinder:
        return {"Ellipsoidal cylinder",
                ":/SampleDesignerToolbox/images/ff_EllipsoidalCylinder_64x64.png",
                "Cylinder with an ellipse cross section"};

    case Type::FullSphere:
        return {"Full sphere", ":/SampleDesignerToolbox/images/ff_FullSphere_64x64.png",
                "Full sphere"};

    case Type::FullSpheroid:
        return {"Full spheroid", ":/SampleDesignerToolbox/images/ff_FullSpheroid_64x64.png",
                "Full spheroid, generated by rotating an ellipse around the vertical axis"};

    case Type::HemiEllipsoid:
        return {"Hemi ellipsoid", ":/SampleDesignerToolbox/images/ff_HemiEllipsoid_64x64.png",
                "A horizontally oriented ellipsoid, truncated at the central plane"};

    case Type::Icosahedron:
        return {"Icosahedron", ":/SampleDesignerToolbox/images/ff_Icosahedron_64x64.png",
                "Icosahedron"};

    case Type::Prism3:
        return {"Prism3", ":/SampleDesignerToolbox/images/ff_Prism3_64x64.png",
                "Prism with an equilateral triangle base"};

    case Type::Prism6:
        return {"Prism6", ":/SampleDesignerToolbox/images/ff_Prism6_64x64.png",
                "Prism with a regular hexagonal base"};

    case Type::Pyramid4:
        return {"Pyramid4", ":/SampleDesignerToolbox/images/ff_Pyramid4_64x64.png",
                "Truncated pyramid with a square base"};

    case Type::CosineRippleBox:
        return {"CosineRippleBox", ":/SampleDesignerToolbox/images/ff_Ripple1_64x64.png",
                "Particle with a cosine profile and a rectangular base"};

    case Type::CosineRippleGauss:
        return {"CosineRippleGauss", ":/SampleDesignerToolbox/images/ff_Ripple1_64x64.png",
                "Particle with a cosine profile and a rectangular base"};

    case Type::CosineRippleLorentz:
        return {"CosineRippleLorentz", ":/SampleDesignerToolbox/images/ff_Ripple1_64x64.png",
                "Particle with a cosine profile and a rectangular base"};

    case Type::SawtoothRippleBox:
        return {"SawtoothRippleBox", ":/SampleDesignerToolbox/images/ff_Ripple2_64x64.png",
                "Particle with an asymmetric triangle profile and a rectangular base"};

    case Type::SawtoothRippleGauss:
        return {"SawtoothRippleGauss", ":/SampleDesignerToolbox/images/ff_Ripple2_64x64.png",
                "Particle with an asymmetric triangle profile and a rectangular base"};

    case Type::SawtoothRippleLorentz:
        return {"SawtoothRippleLorentz", ":/SampleDesignerToolbox/images/ff_Ripple2_64x64.png",
                "Particle with an asymmetric triangle profile and a rectangular base"};

    case Type::Pyramid3:
        return {"Pyramid3", ":/SampleDesignerToolbox/images/ff_Pyramid3_64x64.png",
                "Truncated polyhedron with equilateral triangle base and cropped side faces"};

    case Type::TruncatedCube:
        return {"Truncated cube", ":/SampleDesignerToolbox/images/ff_TruncatedCube_64x64.png",
                "A cube whose eight vertices have been removed"};

    case Type::TruncatedSphere:
        return {"Truncated sphere", ":/SampleDesignerToolbox/images/ff_TruncatedSphere_64x64.png",
                "Spherical dome"};

    case Type::TruncatedSpheroid:
        return {"Truncated spheroid",
                ":/SampleDesignerToolbox/images/ff_TruncatedSpheroid_64x64.png", "Spheroidal dome"};

    default:
        ASSERT(false);
    }
}

FormFactorItemCatalog::Type FormFactorItemCatalog::type(const FormFactorItem* item)
{
    ASSERT(item);

#define CHECK(type)                                                                                \
    if (dynamic_cast<const type##Item*>(item))                                                     \
    return Type::type

    CHECK(Pyramid2);
    CHECK(BarGauss);
    CHECK(BarLorentz);
    CHECK(Box);
    CHECK(Cone);
    CHECK(Pyramid6);
    CHECK(Bipyramid4);
    CHECK(Cylinder);
    CHECK(Dodecahedron);
    CHECK(EllipsoidalCylinder);
    CHECK(FullSphere);
    CHECK(FullSpheroid);
    CHECK(HemiEllipsoid);
    CHECK(Icosahedron);
    CHECK(Prism3);
    CHECK(Prism6);
    CHECK(Pyramid4);
    CHECK(CosineRippleBox);
    CHECK(CosineRippleGauss);
    CHECK(CosineRippleLorentz);
    CHECK(SawtoothRippleBox);
    CHECK(SawtoothRippleGauss);
    CHECK(SawtoothRippleLorentz);
    CHECK(Pyramid3);
    CHECK(TruncatedCube);
    CHECK(TruncatedSphere);
    CHECK(TruncatedSpheroid);
#undef CHECK
    ASSERT(false);
}

QString FormFactorItemCatalog::menuEntry(const FormFactorItem* item)
{
    return uiInfo(type(item)).menuEntry;
}
