//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/CatSample/ItemWithParticlesCatalog.cpp
//! @brief     Implements class ItemWithParticlesCatalog
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/CatSample/ItemWithParticlesCatalog.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"

ItemWithParticles* ItemWithParticlesCatalog::create(Type type, const MaterialItems* materials)
{
    switch (type) {
    case Type::Particle:
        return new ParticleItem(materials);
    case Type::Composition:
        return new ParticleCompositionItem(materials);
    case Type::CoreShell:
        return new ParticleCoreShellItem(materials);
    case Type::MesoCrystal:
        return new MesoCrystalItem(materials);
    default:
        ASSERT(false);
    }
}

QVector<ItemWithParticlesCatalog::Type> ItemWithParticlesCatalog::types()
{
    return {Type::Particle, Type::Composition, Type::CoreShell, Type::MesoCrystal};
}

QVector<ItemWithParticlesCatalog::Type> ItemWithParticlesCatalog::assemblyTypes()
{
    return {Type::Composition, Type::CoreShell, Type::MesoCrystal};
}

ItemWithParticlesCatalog::UiInfo ItemWithParticlesCatalog::uiInfo(Type type)
{
    switch (type) {
    case Type::Particle:
        return {"Particle", "", ""}; // particle is not on UI, only its form factor
    case Type::Composition:
        return {"Particle Composition", "ParticleComposition_64x64.png",
                "Composition of particles with fixed positions"};
    case Type::CoreShell:
        return {"Core shell particle", "ParticleCoreShell_64x64.png",
                "A particle with a core/shell geometry"};
    case Type::MesoCrystal:
        return {"Meso Crystal", "Mesocrystal_64x64.png", "A 3D crystal structure of nanoparticles"};
    default:
        ASSERT(false);
    }
}

ItemWithParticlesCatalog::Type ItemWithParticlesCatalog::type(const ItemWithParticles* item)
{
    ASSERT(item);

    if (dynamic_cast<const ParticleItem*>(item))
        return Type::Particle;
    if (dynamic_cast<const ParticleCompositionItem*>(item))
        return Type::Composition;
    if (dynamic_cast<const MesoCrystalItem*>(item))
        return Type::MesoCrystal;
    if (dynamic_cast<const ParticleCoreShellItem*>(item))
        return Type::CoreShell;

    ASSERT(false);
}
