//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/AxisProperty.h
//! @brief     Defines class AxisProperty
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_AXISPROPERTY_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_AXISPROPERTY_H

#include "Base/Axis/FixedBinAxis.h"
#include "GUI/Model/Descriptor/DoubleProperty.h"
#include "GUI/Model/Descriptor/UIntProperty.h"
#include <memory>

class Streamer;

//! Holds values which can be used to describe a FixedBinAxis.
//!
//! Use this as a member in your class to
//! * reduce dependency to axis items and axis classes
//! * simplify serialization
//! * easily have a UI supporting solution.
//!
//! This class is designed to work seamlessly with the serialization methods; just call it's
//! serialize for both reading and writing.
//! For UI integration just use AxisPropertyEditor.
//!
//! Do not forget to call all three init functions from within the containing class's constructor!

class AxisProperty {
public:
    UINT_PROPERTY(nbins, Nbins);
    DOUBLE_PROPERTY(min, Min);
    DOUBLE_PROPERTY(max, Max);

    void initNbins(const QString& label, const QString& tooltip, uint n = 100,
                   const variant<QString, Unit>& unit = "nbins");

    void initMin(const QString& label, const QString& tooltip, double value,
                 const variant<QString, Unit>& unit, uint decimals = 3,
                 const RealLimits& limits = RealLimits::nonnegative());
    void initMax(const QString& label, const QString& tooltip, double value,
                 const variant<QString, Unit>& unit, uint decimals = 3,
                 const RealLimits& limits = RealLimits::nonnegative());

    void serialize(Streamer& s);

    //! Creates an axis with empty title
    std::unique_ptr<FixedBinAxis> createAxis(double scaleFactor) const;

    void rwAxisProperty(Streamer& s, const QString& tag);
};

//! Add a member and a getter for an axis property
#define AXIS_PROPERTY(nameLower)                                                                   \
protected:                                                                                         \
    AxisProperty m_##nameLower;                                                                    \
                                                                                                   \
public:                                                                                            \
    AxisProperty& nameLower() const { return *const_cast<AxisProperty*>(&m_##nameLower); }

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_AXISPROPERTY_H
