//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/SelectionProperty.h
//! @brief     Defines class SelectionProperty
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_SELECTIONPROPERTY_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_SELECTIONPROPERTY_H

#include "GUI/Model/Descriptor/SelectionDescriptor.h"
#include "GUI/Support/XML/Streamer.h"

// #baMigration docu
template <typename T>
class SelectionProperty {
public:
    ~SelectionProperty() { delete m_p; }

    //! Initializes the selection with the first type in the catalog types.
    //! For initializing a newly created item when creating via Catalog
    template <typename Catalog, typename... ArgsForCreation>
    void init(const QString& label, const QString& tooltip, const QString& persistentTag,
              ArgsForCreation... argsForCreation)
    {
        m_persistentTag = persistentTag;
        initDescriptor<Catalog>(label, tooltip, Catalog::types(), argsForCreation...);
        m_descriptor.setCurrentIndex(0);
    }

    //! Initializes the selection with the first type in the catalog types.
    //! For initializing a newly created item in a separate initializer method
    template <typename Catalog>
    void initWithInitializer(const QString& label, const QString& tooltip,
                             const QString& persistentTag,
                             std::function<void(T, const T)> initializer)
    {
        m_initializer = initializer;
        m_persistentTag = persistentTag;

        initDescriptor<Catalog>(label, tooltip, Catalog::types());
        m_descriptor.setCurrentIndex(0);
    }

    //! Initializes the selection with the first type of the given types subset.
    //! For initializing a newly created item in a separate initializer method.
    //! Only a subset of types shall be used
    template <typename Catalog>
    void initWithInitializer(const QString& label, const QString& tooltip,
                             const QString& persistentTag,
                             const QVector<typename Catalog::Type>& types,
                             std::function<void(T, const T)> initializer)
    {
        m_initializer = initializer;
        m_persistentTag = persistentTag;

        initDescriptor<Catalog>(label, tooltip, types);
        m_descriptor.setCurrentIndex(0);
    }

    SelectionDescriptor<T> descriptor() const { return m_descriptor; }

    operator SelectionDescriptor<T>() const { return m_descriptor; }
    T operator->() const { return m_p; }

    //! Directly set the new item.
    void set(T t, bool callInitializer = false)
    {
        if (callInitializer && m_initializer)
            m_initializer(t, m_p);
        delete m_p;
        m_p = t;
    }

    template <typename S, typename... ArgsForConstructor>
    S* set(ArgsForConstructor... argsForConstructor)
    {
        S* s = new S(argsForConstructor...);
        if (s != nullptr && m_initializer)
            m_initializer(s, m_p);
        m_p = s;
        return s;
    }

    T get() const { return m_p; }

    QString persistentTag() const { return m_persistentTag; }


private:
    template <typename Catalog, typename... ArgsForCreation>
    void initDescriptor(const QString& label, const QString& tooltip,
                        const QVector<typename Catalog::Type>& types,
                        ArgsForCreation... argsForCreation)

    {
        auto setCurrentIndex = [=](int current) {
            auto* p = Catalog::create(types[current], argsForCreation...);
            if (m_initializer)
                m_initializer(p, m_p);
            m_p = p;
        };

        m_descriptor.label = label;
        m_descriptor.tooltip = tooltip;
        for (const auto type : types)
            m_descriptor.options << Catalog::uiInfo(type).menuEntry;
        m_descriptor.currentIndexSetter = setCurrentIndex;
        m_descriptor.currentIndexGetter = [=]() { return types.indexOf(Catalog::type(m_p)); };
        m_descriptor.currentItem = [=] { return m_p; };
    }

private:
    QString m_persistentTag;
    SelectionDescriptor<T> m_descriptor;

    T m_p = nullptr;
    std::function<void(T, const T)> m_initializer;
};

// Extends namespace defined in GUI/Model/XML/Serialize.h
namespace Serialize {

//! Serializes an item from a catalog.
//! Passes optional arguments to the constructor.
template <typename Catalog, typename... ArgsForCreation>
void rwSelected(Streamer& s, SelectionProperty<typename Catalog::CatalogedType*>& d,
                ArgsForCreation... argsForCreation);

} // namespace Serialize

//  ************************************************************************************************
//  Template implementation
//  ************************************************************************************************

template <typename Catalog, typename... ArgsForCreation>
void Serialize::rwSelected(Streamer& s, SelectionProperty<typename Catalog::CatalogedType*>& d,
                           ArgsForCreation... argsForCreation)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        typename Catalog::CatalogedType* p = d.get();
        s.write<Catalog>(d.persistentTag(), p);
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        typename Catalog::CatalogedType* p = nullptr;
        s.read<Catalog>(d.persistentTag(), p, argsForCreation...);
        d.set(p);
    }
}

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_SELECTIONPROPERTY_H
