//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/DetectorItems.cpp
//! @brief     Implements classes DetectorItems
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/DetectorItems.h"
#include "Device/Detector/IDetector2D.h"
#include "Device/Mask/IShape2D.h"
#include "Device/Resolution/ResolutionFunction2DGaussian.h"
#include "GUI/Model/Device/MaskItems.h"
#include "GUI/Model/Device/ResolutionFunctionItems.h"
#include "GUI/Model/Device/SphericalDetectorItem.h"
#include "GUI/Model/Model/SessionModel.h"
#include "GUI/Support/Type/GroupInfo.h"

DetectorItem::DetectorItem() {}

std::unique_ptr<IDetector2D> DetectorItem::createDetector() const
{
    auto result = createDomainDetector();
    addMasksToDomain(result.get());

    if (auto resFunc = createResolutionFunction())
        result->setResolutionFunction(*resFunc);

    return result;
}

void DetectorItem::importMasks(const MaskContainerItem* maskContainer)
{
    m_maskItems.copy(maskContainer);
}

MaskItems& DetectorItem::maskItems()
{
    return m_maskItems;
}

ResolutionFunctionItem* DetectorItem::resolutionFunction() const
{
    return m_resolutionFunction.get();
}

SelectionDescriptor<ResolutionFunctionItem*> DetectorItem::resolutionFunctionSelection() const
{
    return m_resolutionFunction;
}

std::unique_ptr<IResolutionFunction2D> DetectorItem::createResolutionFunction() const
{
    return m_resolutionFunction->createResolutionFunction(axesToDomainUnitsFactor());
}

void DetectorItem::addMasksToDomain(IDetector2D* detector) const
{
    const double scale = axesToDomainUnitsFactor();
    const auto maskItems = m_maskItems.maskItems();
    for (auto maskIter = maskItems.rbegin(); maskIter != maskItems.rend(); maskIter++) {
        if (auto* roiItem = dynamic_cast<RegionOfInterestItem*>(*maskIter)) {
            const double xlow = scale * roiItem->xLow();
            const double ylow = scale * roiItem->yLow();
            const double xup = scale * roiItem->xUp();
            const double yup = scale * roiItem->yUp();
            detector->setRegionOfInterest(xlow, ylow, xup, yup);
        } else {
            std::unique_ptr<IShape2D> shape((*maskIter)->createShape(scale));
            detector->addMask(*shape, (*maskIter)->maskValue());
        }
    }
}
