//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/To/DomainSimulationBuilder.cpp
//! @brief     Implements class DomainSimulationBuilder
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/To/DomainSimulationBuilder.h"
#include "Base/Const/Units.h"
#include "Base/Util/Assert.h"
#include "Core/Background/IBackground.h"
#include "Core/Scan/AlphaScan.h"
#include "Core/Scan/ScanResolution.h"
#include "Core/Simulation/includeSimulations.h"
#include "Device/Beam/Beam.h"
#include "Device/Beam/IFootprintFactor.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Model/Device/BackgroundItems.h"
#include "GUI/Model/Device/BeamAngleItems.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/Model/Device/FootprintItems.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/To/ToDomain.h"
#include "GUI/Support/Data/SimulationOptionsItem.h"
#include "GUI/Util/Error.h"
#include "Param/Distrib/RangedDistributions.h"
#include "Resample/Options/SimulationOptions.h"
#include "Sample/Multilayer/MultiLayer.h"

namespace {

//! If the given item has distributions, then add these distributions to the simulation's
//! distributions handler.
template <class T>
void setParameterDistributionToSimulation(ParameterDistribution::WhichParameter which,
                                          const BeamDistributionItem* item, ISimulation& simulation)
{
    const auto* const parameter_item = dynamic_cast<const T*>(item);
    ASSERT(parameter_item);

    auto P_par_distr = parameter_item->getParameterDistributionForName(which);
    if (P_par_distr)
        simulation.addParameterDistribution(*P_par_distr);
}

//! adds DistributionParameters to the ISimulation
void addDistributionParametersToSimulation(const BeamItem& beam_item, GISASSimulation& simulation)
{
    ASSERT(dynamic_cast<const GISASBeamItem*>(&beam_item));

    setParameterDistributionToSimulation<BeamWavelengthItem>(
        ParameterDistribution::BeamWavelength, beam_item.wavelengthItem(), simulation);
    setParameterDistributionToSimulation<BeamInclinationAngleItem>(
        ParameterDistribution::BeamInclinationAngle, beam_item.inclinationAngleItem(), simulation);
    setParameterDistributionToSimulation<BeamAzimuthalAngleItem>(
        ParameterDistribution::BeamAzimuthalAngle, beam_item.azimuthalAngleItem(), simulation);
}

void addBackgroundToSimulation(const InstrumentItem& instrumentItem, ISimulation& simulation)
{
    auto background = instrumentItem.backgroundItem()->createBackground();
    if (background)
        simulation.setBackground(*background);
}

std::unique_ptr<ScanResolution> createScanResolution(const BeamDistributionItem* beam_item)
{
    if (!beam_item)
        return nullptr;

    const auto* distr_item =
        dynamic_cast<const SymmetricResolutionItem*>(beam_item->distribution());
    if (!distr_item)
        return nullptr;

    const double scale = beam_item->scaleFactor();
    auto ranged_distr = distr_item->createIRangedDistribution(scale);
    if (!ranged_distr)
        return nullptr;

    const double deviation = distr_item->deviation(scale);
    return std::unique_ptr<ScanResolution>(
        ScanResolution::scanAbsoluteResolution(*ranged_distr, deviation));
}

void addBeamDivergencesToScan(const BeamItem& beam_item, AlphaScan& scan)
{
    ASSERT(dynamic_cast<const SpecularBeamItem*>(&beam_item));

    auto resolution = createScanResolution(beam_item.wavelengthItem());
    if (resolution)
        scan.setWavelengthResolution(*resolution);
    resolution = createScanResolution(beam_item.inclinationAngleItem());
    if (resolution)
        scan.setAngleResolution(*resolution);
}

void setSimulationOptions(ISimulation* simulation, const SimulationOptionsItem& optionItem)
{
    simulation->options().setNumberOfThreads(optionItem.numberOfThreads());
    if (optionItem.useMonteCarloIntegration())
        simulation->options().setMonteCarloIntegration(true, optionItem.numberOfMonteCarloPoints());
    simulation->options().setUseAvgMaterials(optionItem.useAverageMaterials());
    simulation->options().setIncludeSpecular(optionItem.includeSpecularPeak());
}

std::unique_ptr<GISASSimulation> createGISASSimulation(std::unique_ptr<MultiLayer> multilayer,
                                                       const GISASInstrumentItem* instrumentItem,
                                                       const SimulationOptionsItem& optionsItem)
{
    std::unique_ptr<const Instrument> instrument = instrumentItem->createInstrument();
    std::unique_ptr<GISASSimulation> result{
        new GISASSimulation(instrument->beam(), *multilayer, instrument->detector())};

    addDistributionParametersToSimulation(*instrumentItem->beamItem(), *result);

    // ISimulation options
    setSimulationOptions(result.get(), optionsItem);

    addBackgroundToSimulation(*instrumentItem, *result);

    return result;
}

std::unique_ptr<OffSpecularSimulation>
createOffSpecularSimulation(std::unique_ptr<MultiLayer> multilayer,
                            const OffSpecularInstrumentItem* instrumentItem,
                            const SimulationOptionsItem& optionsItem)
{
    std::unique_ptr<const Instrument> instrument = instrumentItem->createInstrument();
    std::unique_ptr<OffSpecularSimulation> result{
        new OffSpecularSimulation(instrument->beam(), *multilayer, instrument->detector())};

    auto* beamItem = instrumentItem->beamItem();
    const auto axis = instrumentItem->alphaAxis().createAxis(Units::deg);
    result->setBeamParameters(beamItem->wavelength(), *axis, beamItem->getAzimuthalAngle());

    // TODO Take care about distributions
    // addDistributionParametersToSimulation(*gisasInstrument->beamItem(),
    //                                                          gisas.get());

    // ISimulation options
    setSimulationOptions(result.get(), optionsItem);

    addBackgroundToSimulation(*instrumentItem, *result);

    return result;
}

std::unique_ptr<SpecularSimulation>
createSpecularSimulation(std::unique_ptr<MultiLayer> P_multilayer,
                         const SpecularInstrumentItem* instrument,
                         const SimulationOptionsItem& optionsItem)
{
    std::unique_ptr<SpecularSimulation> result = std::make_unique<SpecularSimulation>();
    result->setSample(*P_multilayer);

    auto* beam_item = instrument->beamItem();
    auto* const axis_item = beam_item->inclinationAxis();
    auto* const footprint = beam_item->footprint();

    AlphaScan scan(beam_item->wavelength(), *axis_item->createAxis(Units::deg));
    scan.setFootprintFactor(footprint->createFootprint().get());

    addBeamDivergencesToScan(*beam_item, scan);

    result->beam().setIntensity(beam_item->intensity());
    result->setScan(scan);

    // ISimulation options
    setSimulationOptions(result.get(), optionsItem);

    addBackgroundToSimulation(*instrument, *result);

    return result;
}

std::unique_ptr<DepthProbeSimulation>
createDepthProbeSimulation(std::unique_ptr<MultiLayer> multilayer,
                           const DepthProbeInstrumentItem* instrument,
                           const SimulationOptionsItem& optionsItem)
{
    std::unique_ptr<DepthProbeSimulation> result = instrument->createSimulation();
    result->setSample(*multilayer);

    setSimulationOptions(result.get(), optionsItem);

    return result;
}

} // namespace


std::unique_ptr<ISimulation>
GUI::Model::DomainSimulationBuilder::createSimulation(const MultiLayerItem* sampleItem,
                                                      const InstrumentItem* instrumentItem,
                                                      const SimulationOptionsItem& optionsItem)
{
    if (sampleItem == nullptr || instrumentItem == nullptr) {
        QString message(
            "GUI::Model::DomainSimulationBuilder::getSimulation() -> Error. Either MultiLayerItem "
            " or InstrumentItem is not defined.");
        throw Error(message);
    }

    auto P_multilayer = GUI::Transform::ToDomain::buildMultiLayer(*sampleItem);

    if (const auto* gisasInstrument = dynamic_cast<const GISASInstrumentItem*>(instrumentItem))
        return createGISASSimulation(std::move(P_multilayer), gisasInstrument, optionsItem);
    if (const auto* offspecInstrument =
            dynamic_cast<const OffSpecularInstrumentItem*>(instrumentItem))
        return createOffSpecularSimulation(std::move(P_multilayer), offspecInstrument, optionsItem);
    if (const auto* specular_instrument =
            dynamic_cast<const SpecularInstrumentItem*>(instrumentItem))
        return createSpecularSimulation(std::move(P_multilayer), specular_instrument, optionsItem);
    if (const auto* penetrator = dynamic_cast<const DepthProbeInstrumentItem*>(instrumentItem))
        return createDepthProbeSimulation(std::move(P_multilayer), penetrator, optionsItem);

    ASSERT(0);
}
