//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Support/XML/Streamer.cpp
//! @brief     Implements class Streamer
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Support/XML/Streamer.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/DeserializationException.h"

void Streamer::writeVersion(unsigned version) const
{
    if (!m_w)
        return;
    m_w->writeAttribute(XML::Tags::Version, QString::number(version));
}

void Streamer::assertVersion(unsigned expectedVersion) const
{
    if (!m_r)
        return;

    const auto foundAsStr = m_r->attributes().value(XML::Tags::Version);
    const unsigned found = foundAsStr.isEmpty() ? 0 : foundAsStr.toUInt();
    if (found < expectedVersion)
        throw DeserializationException::tooOld();
    if (found > expectedVersion)
        throw DeserializationException::tooNew();
}

void Streamer::gotoEndElementOfTag(const QString& tag)
{
    if (m_r->name() != tag) {
        if (!m_r->isEndElement())
            m_r->skipCurrentElement();
        m_r->skipCurrentElement();
    }
    assertCurrentTag(tag);
    if (!m_r->isEndElement())
        m_r->skipCurrentElement();

    assertCurrentToken(QXmlStreamReader::EndElement);
    assertCurrentTag(tag);
}

void Streamer::gotoStartElementOfTag(const QString& tag)
{
    ASSERT(m_r);
    while (m_r->name() != tag && !m_r->hasError())
        m_r->readNextStartElement();
    assertCurrentToken(QXmlStreamReader::StartElement);
}

void Streamer::assertCurrentTag(const QString& expectedTag) const
{
    ASSERT(m_r);

    if (m_r->name() != expectedTag) {
#ifdef _DEBUG
        // to simplify debugging: what is the current tag
        QString foundTag = m_r->name().toString();
        Q_UNUSED(foundTag);
#endif
        throw DeserializationException::streamError();
    }
}

void Streamer::assertCurrentToken(QXmlStreamReader::TokenType token) const
{
    if (m_r->tokenType() != token)
        throw DeserializationException::streamError();
}

void Streamer::start(const QString& tag)
{
    if (m_w)
        m_w->writeStartElement(tag);
    else
        gotoStartElementOfTag(tag);
}

void Streamer::finish(const QString& tag)
{
    if (m_w)
        m_w->writeEndElement();
    else
        gotoEndElementOfTag(tag);
}
