//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/FromDomain/FromDomain.cpp
//! @brief     Implements namespace TransformFromDomain
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/FromDomain/FromDomain.h"
#include "Base/Const/Units.h"
#include "Core/Background/ConstantBackground.h"
#include "Core/Background/PoissonBackground.h"
#include "Core/Scan/AlphaScan.h"
#include "Core/Scan/ScanResolution.h"
#include "Core/Simulation/includeSimulations.h"
#include "Device/Beam/Beam.h"
#include "Device/Beam/FootprintGauss.h"
#include "Device/Beam/FootprintSquare.h"
#include "Device/Detector/RectangularDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "Device/Mask/DetectorMask.h"
#include "Device/Mask/Ellipse.h"
#include "Device/Mask/InfinitePlane.h"
#include "Device/Mask/Line.h"
#include "Device/Mask/Polygon.h"
#include "Device/Mask/Rectangle.h"
#include "Device/Resolution/ConvolutionDetectorResolution.h"
#include "Device/Resolution/ResolutionFunction2DGaussian.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Model/Device/BackgroundItems.h"
#include "GUI/Model/Device/BeamAngleItems.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/MaskItems.h"
#include "GUI/Model/Device/RectangularDetectorItem.h"
#include "GUI/Model/Device/ResolutionFunctionItems.h"
#include "GUI/Model/Device/SphericalDetectorItem.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Util/Error.h"
#include "Param/Distrib/Distributions.h"
#include "Param/Distrib/RangedDistributions.h"
#include "Param/Node/NodeUtils.h"
#include "Sample/Aggregate/Interferences.h"
#include "Sample/HardParticle/HardParticles.h"
#include "Sample/Interface/LayerInterface.h"
#include "Sample/Interface/LayerRoughness.h"
#include "Sample/Multilayer/Layer.h"
#include "Sample/Scattering/Rotations.h"

using namespace NodeUtils;

namespace {

void setPDF1D(InterferenceRadialParaCrystalItem* item, const IProfile1D* ipdf)
{
    if (const auto* pdf = dynamic_cast<const Profile1DCauchy*>(ipdf)) {
        auto* pdfItem = new Profile1DCauchyItem();
        pdfItem->omega().set(pdf->omega());
        item->setPDFType(pdfItem);
    } else if (const auto* pdf = dynamic_cast<const Profile1DGauss*>(ipdf)) {
        auto* pdfItem = new Profile1DGaussItem();
        pdfItem->omega().set(pdf->omega());
        item->setPDFType(pdfItem);
    } else if (const auto* pdf = dynamic_cast<const Profile1DGate*>(ipdf)) {
        auto* pdfItem = new Profile1DGateItem();
        pdfItem->omega().set(pdf->omega());
        item->setPDFType(pdfItem);
    } else if (const auto* pdf = dynamic_cast<const Profile1DTriangle*>(ipdf)) {
        auto* pdfItem = new Profile1DTriangleItem();
        pdfItem->omega().set(pdf->omega());
        item->setPDFType(pdfItem);
    } else if (const auto* pdf = dynamic_cast<const Profile1DCosine*>(ipdf)) {
        auto* pdfItem = new Profile1DCosineItem();
        pdfItem->omega().set(pdf->omega());
        item->setPDFType(pdfItem);
    } else if (const auto* pdf = dynamic_cast<const Profile1DVoigt*>(ipdf)) {
        auto* pdfItem = new Profile1DVoigtItem();
        pdfItem->omega().set(pdf->omega());
        pdfItem->eta().set(pdf->eta());
        item->setPDFType(pdfItem);
    } else
        throw Error("GUI::Transform::FromDomain::setPDF1D: -> Error");
}

// note: SetterPDF(1|2)Type are needed because template template parameter must be classes

template <class T>
struct SetterPDF1Type {
    T* operator()(Interference2DParaCrystalItem* item)
    {
        auto* p = new T();
        item->setPDF1Type(p);
        return p;
    }
};

template <class T>
struct SetterPDF2Type {
    T* operator()(Interference2DParaCrystalItem* item)
    {
        auto* p = new T();
        item->setPDF2Type(p);
        return p;
    }
};

template <template <class T> class setType>
void setPDF2D(Interference2DParaCrystalItem* item, const IProfile2D* pdf)
{
    if (const auto* pdf_cauchy = dynamic_cast<const Profile2DCauchy*>(pdf)) {
        Profile2DCauchyItem* pdfItem = setType<Profile2DCauchyItem>()(item);
        pdfItem->setOmegaX(pdf_cauchy->omegaX());
        pdfItem->setOmegaY(pdf_cauchy->omegaY());
        pdfItem->setGamma(Units::rad2deg(pdf_cauchy->gamma()));
    } else if (const auto* pdf_gauss = dynamic_cast<const Profile2DGauss*>(pdf)) {
        Profile2DGaussItem* pdfItem = setType<Profile2DGaussItem>()(item);
        pdfItem->setOmegaX(pdf_gauss->omegaX());
        pdfItem->setOmegaY(pdf_gauss->omegaY());
        pdfItem->setGamma(Units::rad2deg(pdf_gauss->gamma()));
    } else if (const auto* pdf_gate = dynamic_cast<const Profile2DGate*>(pdf)) {
        Profile2DGateItem* pdfItem = setType<Profile2DGateItem>()(item);
        pdfItem->setOmegaX(pdf_gate->omegaX());
        pdfItem->setOmegaY(pdf_gate->omegaY());
        pdfItem->setGamma(Units::rad2deg(pdf_gate->gamma()));
    } else if (const auto* pdf_cone = dynamic_cast<const Profile2DCone*>(pdf)) {
        Profile2DConeItem* pdfItem = setType<Profile2DConeItem>()(item);
        pdfItem->setOmegaX(pdf_cone->omegaX());
        pdfItem->setOmegaY(pdf_cone->omegaY());
        pdfItem->setGamma(Units::rad2deg(pdf_cone->gamma()));
    } else if (const auto* pdf_voigt = dynamic_cast<const Profile2DVoigt*>(pdf)) {
        Profile2DVoigtItem* pdfItem = setType<Profile2DVoigtItem>()(item);
        pdfItem->setOmegaX(pdf_voigt->omegaX());
        pdfItem->setOmegaY(pdf_voigt->omegaY());
        pdfItem->setGamma(Units::rad2deg(pdf_voigt->gamma()));
        pdfItem->setEta(pdf_voigt->eta());
    } else
        throw Error("GUI::Transform::FromDomain::setPDF2D: -> Error");
}

void setDecayFunction1D(Interference1DLatticeItem* item, const IProfile1D* ipdf)
{
    if (const auto* pdf = dynamic_cast<const Profile1DCauchy*>(ipdf)) {
        auto* pdfItem = new Profile1DCauchyItem();
        pdfItem->omega().set(pdf->omega());
        item->setDecayFunction(pdfItem);
    } else if (const auto* pdf = dynamic_cast<const Profile1DGauss*>(ipdf)) {
        auto* pdfItem = new Profile1DGaussItem();
        pdfItem->omega().set(pdf->omega());
        item->setDecayFunction(pdfItem);
    } else if (const auto* pdf = dynamic_cast<const Profile1DTriangle*>(ipdf)) {
        auto* pdfItem = new Profile1DTriangleItem();
        pdfItem->omega().set(pdf->omega());
        item->setDecayFunction(pdfItem);
    } else if (const auto* pdf = dynamic_cast<const Profile1DVoigt*>(ipdf)) {
        auto* pdfItem = new Profile1DVoigtItem();
        pdfItem->omega().set(pdf->omega());
        pdfItem->eta().set(pdf->eta());
        item->setDecayFunction(pdfItem);
    } else
        throw Error("GUI::Transform::FromDomain::setDecayFunction1D: -> Error");
}

void setDecayFunction2D(Interference2DLatticeItem* item, const IProfile2D* pdf)
{
    if (const auto* pdf_cauchy = dynamic_cast<const Profile2DCauchy*>(pdf)) {
        auto* pdfItem = new Profile2DCauchyItem();
        pdfItem->setOmegaX(pdf_cauchy->omegaX());
        pdfItem->setOmegaY(pdf_cauchy->omegaY());
        pdfItem->setGamma(Units::rad2deg(pdf_cauchy->gamma()));
        item->setDecayFunctionType(pdfItem);
    } else if (const auto* pdf_gauss = dynamic_cast<const Profile2DGauss*>(pdf)) {
        auto* pdfItem = new Profile2DGaussItem();
        pdfItem->setOmegaX(pdf_gauss->omegaX());
        pdfItem->setOmegaY(pdf_gauss->omegaY());
        pdfItem->setGamma(Units::rad2deg(pdf_gauss->gamma()));
        item->setDecayFunctionType(pdfItem);
    } else if (const auto* pdf_voigt = dynamic_cast<const Profile2DVoigt*>(pdf)) {
        auto* pdfItem = new Profile2DVoigtItem();
        pdfItem->setOmegaX(pdf_voigt->omegaX());
        pdfItem->setOmegaY(pdf_voigt->omegaY());
        pdfItem->setGamma(Units::rad2deg(pdf_voigt->gamma()));
        pdfItem->setEta(pdf_voigt->eta());
        item->setDecayFunctionType(pdfItem);
    } else
        throw Error("GUI::Transform::FromDomain::setDecayFunction2D: -> Error");
}

void set2DLatticeParameters(Interference2DAbstractLatticeItem* item, const Lattice2D& lattice)
{
    Lattice2DItem* latticeItem(nullptr);
    if (lattice.className() == "SquareLattice2D") {
        auto* squareLatticeItem = new SquareLattice2DItem();
        squareLatticeItem->setLatticeLength(lattice.length1());
        latticeItem = squareLatticeItem;
    } else if (lattice.className() == "HexagonalLattice2D") {
        auto* hexLatticeItem = new HexagonalLattice2DItem();
        hexLatticeItem->setLatticeLength(lattice.length1());
        latticeItem = hexLatticeItem;
    } else {
        auto* basicLatticeItem = new BasicLattice2DItem();
        basicLatticeItem->setLatticeLength1(lattice.length1());
        basicLatticeItem->setLatticeLength2(lattice.length2());
        basicLatticeItem->setLatticeAngle(Units::rad2deg(lattice.latticeAngle()));
        latticeItem = basicLatticeItem;
    }
    latticeItem->setLatticeRotationAngle(Units::rad2deg(lattice.rotationAngle()));
    item->setLatticeType(latticeItem);
}

void setPositionVariance(InterferenceItem* item, const IInterference& iff)
{
    double pos_var = iff.positionVariance();
    item->positionVariance().set(pos_var);
}

void setDistribution(BeamDistributionItem* part_distr_item, ParameterDistribution par_distr,
                     double factor)
{
    const IDistribution1D* p_distribution = par_distr.getDistribution();

    DistributionItem* distItem = nullptr;
    if (const auto* distr = dynamic_cast<const DistributionGate*>(p_distribution)) {
        auto* distr_gate_item = part_distr_item->setDistributionType<DistributionGateItem>();
        distr_gate_item->setRange(factor * distr->lowerBound(), factor * distr->upperBound());
        distItem = distr_gate_item;
    } else if (const auto* distr = dynamic_cast<const DistributionLorentz*>(p_distribution)) {
        auto* distr_lorentz_item = part_distr_item->setDistributionType<DistributionLorentzItem>();
        distr_lorentz_item->setMean(factor * distr->mean());
        distr_lorentz_item->setHwhm(factor * distr->hwhm());
        distItem = distr_lorentz_item;
    } else if (const auto* distr = dynamic_cast<const DistributionGaussian*>(p_distribution)) {
        auto* distr_gauss_item = part_distr_item->setDistributionType<DistributionGaussianItem>();
        distr_gauss_item->setMean(factor * distr->mean());
        distr_gauss_item->setStandardDeviation(factor * distr->getStdDev());
        distItem = distr_gauss_item;
    } else if (const auto* distr = dynamic_cast<const DistributionLogNormal*>(p_distribution)) {
        auto* distr_lognorm_item =
            part_distr_item->setDistributionType<DistributionLogNormalItem>();
        distr_lognorm_item->setMedian(factor * distr->getMedian());
        distr_lognorm_item->setScaleParameter(distr->getScalePar());
        distItem = distr_lognorm_item;
    } else if (const auto* distr = dynamic_cast<const DistributionCosine*>(p_distribution)) {
        auto* distr_cos_item = part_distr_item->setDistributionType<DistributionCosineItem>();
        distr_cos_item->setMean(factor * distr->mean());
        distr_cos_item->setSigma(factor * distr->sigma());
        distItem = distr_cos_item;
    } else if (const auto* distr = dynamic_cast<const DistributionTrapezoid*>(p_distribution)) {
        auto* distr_trapez_item = part_distr_item->setDistributionType<DistributionTrapezoidItem>();
        distr_trapez_item->setCenter(factor * distr->mean());
        distr_trapez_item->setLeftWidth(factor * distr->getLeftWidth());
        distr_trapez_item->setMiddleWidth(factor * distr->getMiddleWidth());
        distr_trapez_item->setRightWidth(factor * distr->getRightWidth());
        distItem = distr_trapez_item;
    } else
        throw Error("GUI::Transform::FromDomain::setDistribution: -> unknown distribution");

    distItem->setNumberOfSamples((int)par_distr.getNbrSamples());

    if (distItem->hasSigmaFactor())
        distItem->setSigmaFactor(par_distr.sigmaFactor());

    // TODO It's wrong if domain distribution made for angles.
    distItem->setLimits(par_distr.getLimits().scaledLimits(factor));
}

void addDistributionToBeamItem(ParameterDistribution::WhichParameter which,
                               BeamDistributionItem* item,
                               const ParameterDistribution& distribution)
{
    if (distribution.whichParameter() != which)
        return;

    GUI::Transform::FromDomain::setItemFromSample(item, distribution);
}

void addRangedDistributionToItem(BeamDistributionItem* distributionItem,
                                 const IRangedDistribution& ranged, double mean, double std_dev)
{
    if (!distributionItem)
        return;
    ParameterDistribution par_distr(ParameterDistribution::None,
                                    *ranged.distribution(mean, std_dev), ranged.nSamples(),
                                    ranged.sigmaFactor(), ranged.limits());
    GUI::Transform::FromDomain::setItemFromSample(distributionItem, par_distr);
}

template <typename T>
T* addFormFactorItem(std::variant<ParticleItem*, MesoCrystalItem*> parent)
{
    if (std::holds_alternative<ParticleItem*>(parent))
        return std::get<ParticleItem*>(parent)->setFormFactorType<T>();

    return std::get<MesoCrystalItem*>(parent)->setOuterShapeType<T>();
}

} // namespace

//  ************************************************************************************************
//  class implementation
//  ************************************************************************************************

void GUI::Transform::FromDomain::set1DLatticeItem(Interference1DLatticeItem* item,
                                                  const Interference1DLattice& sample)
{
    item->length().set(sample.length());
    item->rotationAngle().set(Units::rad2deg(sample.xi()));

    const auto* pdf = OnlyChildOfType<IProfile1D>(sample);
    setDecayFunction1D(item, pdf);
    setPositionVariance(item, sample);
}

void GUI::Transform::FromDomain::set2DLatticeItem(Interference2DLatticeItem* item,
                                                  const Interference2DLattice& sample)
{
    set2DLatticeParameters(item, sample.lattice());

    item->setXiIntegration(sample.integrationOverXi());

    const auto* p_pdf = OnlyChildOfType<IProfile2D>(sample);
    setDecayFunction2D(item, p_pdf);
    setPositionVariance(item, sample);
}

void GUI::Transform::FromDomain::set2DParaCrystalItem(Interference2DParaCrystalItem* item,
                                                      const Interference2DParaCrystal& sample)
{
    set2DLatticeParameters(item, sample.lattice());

    item->setDampingLength(sample.dampingLength());
    item->setDomainSize1(sample.domainSizes()[0]);
    item->setDomainSize2(sample.domainSizes()[1]);
    item->setXiIntegration(sample.integrationOverXi());

    auto pdfs = ChildNodesOfType<IProfile2D>(sample);

    if (!pdfs.empty()) {
        setPDF2D<SetterPDF1Type>(item, pdfs[0]);
        if (pdfs.size() >= 2)
            setPDF2D<SetterPDF2Type>(item, pdfs[1]);
    }
    setPositionVariance(item, sample);
}

void GUI::Transform::FromDomain::setFinite2DLatticeItem(InterferenceFinite2DLatticeItem* item,
                                                        const InterferenceFinite2DLattice& sample)
{
    set2DLatticeParameters(item, sample.lattice());

    item->setDomainSize1(sample.numberUnitCells1());
    item->setDomainSize2(sample.numberUnitCells2());
    item->setXiIntegration(sample.integrationOverXi());
    setPositionVariance(item, sample);
}

void GUI::Transform::FromDomain::setHardDiskItem(InterferenceHardDiskItem* item,
                                                 const InterferenceHardDisk& sample)
{
    item->radius().set(sample.radius());
    item->density().set(sample.density());
    setPositionVariance(item, sample);
}

void GUI::Transform::FromDomain::setRadialParaCrystalItem(
    InterferenceRadialParaCrystalItem* item, const InterferenceRadialParaCrystal& sample)
{
    item->peakDistance().set(sample.peakDistance());
    item->dampingLength().set(sample.dampingLength());
    item->domainSize().set(sample.domainSize());
    item->kappa().set(sample.kappa());

    const auto* ipdf = OnlyChildOfType<IProfile1D>(sample);
    setPDF1D(item, ipdf);
    setPositionVariance(item, sample);
}

void GUI::Transform::FromDomain::setLayerItem(LayerItem* layerItem, const Layer* layer,
                                              const LayerInterface* top_interface)
{
    layerItem->thickness().set(layer->thickness());
    layerItem->clearRoughness();
    layerItem->numSlices().set(layer->numberOfSlices());

    if (top_interface) {
        const LayerRoughness* roughness = top_interface->roughness();
        if (GUI::Transform::FromDomain::isValidRoughness(roughness)) {
            layerItem->setBasicRoughness();
            GUI::Transform::FromDomain::setRoughnessItem(layerItem->roughness().currentItem(),
                                                         *roughness);
        }
    }
}

void GUI::Transform::FromDomain::setRoughnessItem(LayerBasicRoughnessItem* item,
                                                  const LayerRoughness& sample)
{
    item->sigma().set(sample.sigma());
    item->hurst().set(sample.hurst());
    item->lateralCorrelationLength().set(sample.lateralCorrLength());
}

//! Returns true if given roughness is non-zero roughness
bool GUI::Transform::FromDomain::isValidRoughness(const LayerRoughness* roughness)
{
    if (!roughness)
        return false;
    if (roughness->sigma() == 0 && roughness->hurst() == 0.0
        && roughness->lateralCorrLength() == 0.0)
        return false;
    return true;
}

void GUI::Transform::FromDomain::setGISASBeamItem(BeamItem* beam_item,
                                                  const GISASSimulation& simulation)
{
    ASSERT(beam_item);
    const Beam& beam = simulation.beam();

    beam_item->setIntensity(beam.intensity());
    beam_item->setWavelength(beam.wavelength());
    beam_item->setInclinationAngle(Units::rad2deg(beam.direction().alpha()));
    beam_item->setAzimuthalAngle(Units::rad2deg(beam.direction().phi()));

    // distribution parameters
    const std::vector<ParameterDistribution>& distributions = simulation.getDistributions();
    for (size_t i = 0; i < distributions.size(); ++i) {
        // try all, one will have the right 'whichParameter'
        addDistributionToBeamItem(ParameterDistribution::BeamWavelength,
                                  beam_item->wavelengthItem(), distributions[i]);
        addDistributionToBeamItem(ParameterDistribution::BeamInclinationAngle,
                                  beam_item->inclinationAngleItem(), distributions[i]);
        addDistributionToBeamItem(ParameterDistribution::BeamAzimuthalAngle,
                                  beam_item->azimuthalAngleItem(), distributions[i]);
    }
}

void GUI::Transform::FromDomain::setOffSpecularBeamItem(BeamItem* beam_item,
                                                        const OffSpecularSimulation& simulation)
{
    const Beam& beam = simulation.beam();

    beam_item->setIntensity(beam.intensity());
    beam_item->setWavelength(beam.wavelength());
    beam_item->setInclinationAngle(Units::rad2deg(beam.direction().alpha()));
    beam_item->setAzimuthalAngle(Units::rad2deg(beam.direction().phi()));
    // TODO implement beam divergence
}

void GUI::Transform::FromDomain::setSpecularBeamItem(SpecularBeamItem* beam_item,
                                                     const SpecularSimulation& simulation)
{
    const Beam& beam = simulation.beam();

    beam_item->setIntensity(beam.intensity());
    beam_item->setWavelength(beam.wavelength());
    beam_item->setInclinationAngle(0.0); // inclination angle is hardcoded
    beam_item->setAzimuthalAngle(0.0);   // azimuthal angle is hardcoded

    auto* axis_item = beam_item->inclinationAxis();
    GUI::Transform::FromDomain::setAxisItem(axis_item, *simulation.coordinateAxis(),
                                            1. / Units::deg);

    setFootprintFactor(simulation.footprintFactor(), beam_item);

    const auto* scan = dynamic_cast<const AlphaScan*>(simulation.dataHandler());
    if (!scan)
        return;

    const auto* resolution = scan->wavelengthResolution();
    if (!resolution->empty()) {
        double mean = scan->wavelength();
        double std_dev = resolution->stdDevs(mean, 1).front();
        addRangedDistributionToItem(beam_item->wavelengthItem(), *resolution->distribution(), mean,
                                    std_dev);
    }

    resolution = scan->angleResolution();
    if (resolution && !resolution->empty()) {
        double std_dev = resolution->stdDevs(0.0, 1).front();
        addRangedDistributionToItem(beam_item->inclinationAngleItem(), *resolution->distribution(),
                                    0.0, std_dev);
    }
}

void GUI::Transform::FromDomain::setDetector(Instrument2DItem* instrument_item,
                                             const ISimulation2D& simulation)
{
    const IDetector& detector = simulation.detector();
    setDetectorGeometry(instrument_item, detector);

    auto* detector_item = instrument_item->detector();

    setDetectorResolution(detector_item, detector);
    setDetectorMasks(detector_item, simulation);
    setPolarizerAnalyzer(instrument_item, simulation);
}

void GUI::Transform::FromDomain::setDetectorGeometry(Instrument2DItem* instrument_item,
                                                     const IDetector& detector)
{
    if (const auto* det = dynamic_cast<const SphericalDetector*>(&detector)) {
        auto* item = instrument_item->setDetectorType<SphericalDetectorItem>();
        setSphericalDetector(item, *det);
    } else if (const auto* det = dynamic_cast<const RectangularDetector*>(&detector)) {
        auto* item = instrument_item->setDetectorType<RectangularDetectorItem>();
        setRectangularDetector(item, *det);
    } else {
        throw Error("GUI::Transform::FromDomain::setDetectorGeometry() -> Unknown detector type.");
    }
}

void GUI::Transform::FromDomain::setDetectorResolution(DetectorItem* detector_item,
                                                       const IDetector& detector)
{
    const IDetectorResolution* p_resfunc = detector.detectorResolution();

    if (!p_resfunc)
        return;

    if (const auto* p_convfunc = dynamic_cast<const ConvolutionDetectorResolution*>(p_resfunc)) {
        if (const auto* resfunc = dynamic_cast<const ResolutionFunction2DGaussian*>(
                p_convfunc->getResolutionFunction2D())) {
            auto* item =
                detector_item->setResolutionFunctionType<ResolutionFunction2DGaussianItem>();
            const double scale = 1.0 / detector_item->axesToDomainUnitsFactor();
            item->setSigmaX(scale * resfunc->sigmaX());
            item->setSigmaY(scale * resfunc->sigmaY());
        } else {
            throw Error("GUI::Transform::FromDomain::setDetectorResolution() -> Error. "
                        "Unknown detector resolution function");
        }
    } else {
        throw Error("GUI::Transform::FromDomain::setDetectorResolution() -> Error. "
                    "Not a ConvolutionDetectorResolution function");
    }
}

void GUI::Transform::FromDomain::setPolarizerAnalyzer(Instrument2DItem* instrument_item,
                                                      const ISimulation2D& simulation)
{
    instrument_item->setPolarization(simulation.beam().getBlochVector());

    const IDetector& detector = simulation.detector();
    double total_transmission = detector.analyzer().totalTransmission();
    if (total_transmission <= 0.0)
        return;

    R3 analyzer_dir = detector.analyzer().polDirection();
    double efficiency = detector.analyzer().polEfficiency();
    instrument_item->setAnalyzerDirection(analyzer_dir);
    instrument_item->setAnalyzerEfficiency(efficiency);
    instrument_item->setAnalyzerTotalTransmission(total_transmission);
}

void GUI::Transform::FromDomain::setSphericalDetector(SphericalDetectorItem* detectorItem,
                                                      const SphericalDetector& detector)
{
    // Axes
    const IAxis& phi_axis = detector.axis(0);
    const IAxis& alpha_axis = detector.axis(1);

    auto& phiAxis = detectorItem->phiAxis();
    phiAxis.setNbins(phi_axis.size());
    phiAxis.setMin(Units::rad2deg(phi_axis.lowerBound()));
    phiAxis.setMax(Units::rad2deg(phi_axis.upperBound()));

    auto& alphaAxis = detectorItem->alphaAxis();
    alphaAxis.setNbins(alpha_axis.size());
    alphaAxis.setMin(Units::rad2deg(alpha_axis.lowerBound()));
    alphaAxis.setMax(Units::rad2deg(alpha_axis.upperBound()));
}

void GUI::Transform::FromDomain::setRectangularDetector(RectangularDetectorItem* detectorItem,
                                                        const RectangularDetector& detector)
{
    // Axes
    detectorItem->setXSize(detector.getNbinsX());
    detectorItem->setWidth(detector.width());

    detectorItem->setYSize(detector.getNbinsY());
    detectorItem->setHeight(detector.height());

    detectorItem->setDetectorAlignment(detector.getDetectorArrangment());
    if (detector.getDetectorArrangment() == RectangularDetector::GENERIC) {
        R3 normal = detector.getNormalVector();
        detectorItem->setNormalVector(normal);

        R3 direction = detector.getDirectionVector();
        detectorItem->setDirectionVector(direction);

        detectorItem->setU0(detector.getU0());
        detectorItem->setV0(detector.getV0());
    }

    else if (detector.getDetectorArrangment() == RectangularDetector::PERPENDICULAR_TO_SAMPLE) {
        detectorItem->setDistance(detector.getDistance());
        detectorItem->setU0(detector.getU0());
        detectorItem->setV0(detector.getV0());
    } else if (detector.getDetectorArrangment()
               == RectangularDetector::PERPENDICULAR_TO_DIRECT_BEAM) {
        detectorItem->setDistance(detector.getDistance());
        detectorItem->setU0(detector.getU0());
        detectorItem->setV0(detector.getV0());
    } else if (detector.getDetectorArrangment()
               == RectangularDetector::PERPENDICULAR_TO_REFLECTED_BEAM) {
        detectorItem->setDistance(detector.getDistance());
        detectorItem->setU0(detector.getU0());
        detectorItem->setV0(detector.getV0());
    } else if (detector.getDetectorArrangment()
               == RectangularDetector::PERPENDICULAR_TO_REFLECTED_BEAM_DPOS) {
        detectorItem->setDistance(detector.getDistance());
        detectorItem->setU0(detector.getDirectBeamU0());
        detectorItem->setV0(detector.getDirectBeamV0());
    } else {
        throw Error("GUI::Transform::FromDomain::setItemFromSample(RectangularDetectorItem* "
                    "detectorItem "
                    "Error. Unknown detector arrangement");
    }
}

void GUI::Transform::FromDomain::setDetectorMasks(DetectorItem* detector_item,
                                                  const ISimulation& simulation)
{
    const IDetector& detector = simulation.detector();
    if ((detector.detectorMask() && detector.detectorMask()->hasMasks())
        || detector.hasExplicitRegionOfInterest()) {
        const double scale = 1.0 / detector_item->axesToDomainUnitsFactor();
        setMaskContainer(&detector_item->maskItems(), detector, scale);
    }
}

void GUI::Transform::FromDomain::setMaskContainer(MaskItems* destMaskItems,
                                                  const IDetector& detector, double scale)
{
    const auto* detectorMask = detector.detectorMask();
    for (size_t i_mask = 0; i_mask < detectorMask->numberOfMasks(); ++i_mask) {
        bool mask_value(false);
        const IShape2D* shape = detectorMask->getMaskShape(i_mask, mask_value);

        if (const auto* ellipse = dynamic_cast<const Ellipse*>(shape)) {
            auto* ellipseItem = new EllipseItem();
            ellipseItem->setXCenter(scale * ellipse->getCenterX());
            ellipseItem->setYCenter(scale * ellipse->getCenterY());
            ellipseItem->setXRadius(scale * ellipse->radiusX());
            ellipseItem->setYRadius(scale * ellipse->radiusY());
            ellipseItem->setAngle(scale * ellipse->getTheta());
            ellipseItem->setMaskValue(mask_value);
            destMaskItems->insertMask(0, ellipseItem);
        }

        else if (const auto* rectangle = dynamic_cast<const Rectangle*>(shape)) {
            auto* rectangleItem = new RectangleItem();
            rectangleItem->setXLow(scale * rectangle->getXlow());
            rectangleItem->setYLow(scale * rectangle->getYlow());
            rectangleItem->setXUp(scale * rectangle->getXup());
            rectangleItem->setYUp(scale * rectangle->getYup());
            rectangleItem->setMaskValue(mask_value);
            destMaskItems->insertMask(0, rectangleItem);
        }

        else if (const auto* polygon = dynamic_cast<const Polygon*>(shape)) {
            auto* polygonItem = new PolygonItem();
            std::vector<double> xpos, ypos;
            polygon->getPoints(xpos, ypos);
            for (size_t i_point = 0; i_point < xpos.size(); ++i_point)
                polygonItem->addPoint(scale * xpos[i_point], scale * ypos[i_point]);

            polygonItem->setMaskValue(mask_value);
            polygonItem->setIsClosed(true);

            destMaskItems->insertMask(0, polygonItem);
        }

        else if (const auto* vline = dynamic_cast<const VerticalLine*>(shape)) {
            auto* lineItem = new VerticalLineItem();
            lineItem->setPosX(scale * vline->getXpos());
            lineItem->setMaskValue(mask_value);
            destMaskItems->insertMask(0, lineItem);
        }

        else if (const auto* hline = dynamic_cast<const HorizontalLine*>(shape)) {
            auto* lineItem = new HorizontalLineItem();
            lineItem->setPosY(scale * hline->getYpos());
            lineItem->setMaskValue(mask_value);
            destMaskItems->insertMask(0, lineItem);
        }

        else if (const auto* plane = dynamic_cast<const InfinitePlane*>(shape)) {
            Q_UNUSED(plane);
            auto* planeItem = new MaskAllItem();
            planeItem->setMaskValue(mask_value);
            destMaskItems->insertMask(-1, planeItem);
        }

        else {
            throw Error("GUI::Transform::FromDomain::setDetectorMasks() -> Error. "
                        "Unknown shape");
        }
    }

    if (detector.hasExplicitRegionOfInterest()) {
        const auto xBounds = detector.regionOfInterestBounds(0);
        const auto yBounds = detector.regionOfInterestBounds(1);

        auto* roiItem = new RegionOfInterestItem();
        roiItem->setXLow(scale * xBounds.first);
        roiItem->setYLow(scale * yBounds.first);
        roiItem->setXUp(scale * xBounds.second);
        roiItem->setYUp(scale * yBounds.second);
        destMaskItems->insertMask(-1, roiItem);
    }
}

void GUI::Transform::FromDomain::setItemFromSample(
    BeamDistributionItem* beam_distribution_item,
    const ParameterDistribution& parameter_distribution)
{
    ASSERT(beam_distribution_item);

    if (parameter_distribution.getMinValue() < parameter_distribution.getMaxValue()) {
        throw Error(
            "GUI::Transform::FromDomain::setItemFromSample(BeamDistributionItem* "
            "beamDistributionItem,"
            "const ParameterDistribution& parameterDistribution) -> Error. ParameterDistribution "
            "with defined min,max are not yet implemented in GUI");
    }

    const double unit_factor = 1.0 / beam_distribution_item->scaleFactor();
    setDistribution(beam_distribution_item, parameter_distribution, unit_factor);
}

void GUI::Transform::FromDomain::setBackground(InstrumentItem* instrument_item,
                                               const ISimulation& simulation)
{
    const auto* p_bg = simulation.background();
    if (const auto* p_constant_bg = dynamic_cast<const ConstantBackground*>(p_bg)) {
        auto* constant_bg_item = instrument_item->setBackgroundType<ConstantBackgroundItem>();
        double value = p_constant_bg->backgroundValue();
        constant_bg_item->setBackgroundValue(value);
    } else if (dynamic_cast<const PoissonBackground*>(p_bg))
        instrument_item->setBackgroundType<PoissonBackgroundItem>();
}

void GUI::Transform::FromDomain::setFootprintFactor(const IFootprintFactor* footprint,
                                                    SpecularBeamItem* beam_item)
{
    if (!footprint)
        return;
    if (const auto* const gaussian_fp = dynamic_cast<const FootprintGauss*>(footprint))
        beam_item->setGaussianFootprint(gaussian_fp->widthRatio());
    else if (const auto* const square_fp = dynamic_cast<const FootprintSquare*>(footprint))
        beam_item->setSquareFootprint(square_fp->widthRatio());
}

void GUI::Transform::FromDomain::setAxisItem(BasicAxisItem* item, const IAxis& axis, double factor)
{
    if (!dynamic_cast<const FixedBinAxis*>(&axis))
        throw Error("GUI::Transform::FromDomain::setAxisItem() -> Error. Unexpected axis");

    item->setBinCount(static_cast<int>(axis.size()));
    item->setLowerBound(factor * axis.lowerBound());
    item->setUpperBound(factor * axis.upperBound());
    item->setTitle(QString::fromStdString(axis.axisName()));
}

void GUI::Transform::FromDomain::setRotation(ItemWithParticles* item, const IRotation* rotation)
{
    if (!rotation)
        item->setRotation(nullptr);
    else if (const auto* r = dynamic_cast<const RotationX*>(rotation)) {
        auto* rotationItem = new XRotationItem();
        rotationItem->angle().set(Units::rad2deg(r->angle()));
        item->setRotation(rotationItem);
    } else if (const auto* r = dynamic_cast<const RotationY*>(rotation)) {
        auto* rotationItem = new YRotationItem();
        rotationItem->angle().set(Units::rad2deg(r->angle()));
        item->setRotation(rotationItem);
    } else if (const auto* r = dynamic_cast<const RotationZ*>(rotation)) {
        auto* rotationItem = new ZRotationItem();
        rotationItem->angle().set(Units::rad2deg(r->angle()));
        item->setRotation(rotationItem);
    } else if (const auto* r = dynamic_cast<const RotationEuler*>(rotation)) {
        auto* rotationItem = new EulerRotationItem();
        rotationItem->alpha().set(Units::rad2deg(r->alpha()));
        rotationItem->beta().set(Units::rad2deg(r->beta()));
        rotationItem->gamma().set(Units::rad2deg(r->gamma()));
        item->setRotation(rotationItem);
    }
}

void GUI::Transform::FromDomain::setInterference(ParticleLayoutItem* layoutItem,
                                                 const IInterference* interference)
{
    if (!interference)
        layoutItem->removeInterference();
    else if (const auto* itf = dynamic_cast<const Interference1DLattice*>(interference)) {
        auto* itfItem = new Interference1DLatticeItem();
        set1DLatticeItem(itfItem, *itf);
        layoutItem->setInterference(itfItem);
    } else if (const auto* itf = dynamic_cast<const Interference2DLattice*>(interference)) {
        auto* itfItem = new Interference2DLatticeItem();
        set2DLatticeItem(itfItem, *itf);
        layoutItem->setInterference(itfItem);
    } else if (const auto* itf = dynamic_cast<const Interference2DParaCrystal*>(interference)) {
        auto* itfItem = new Interference2DParaCrystalItem();
        set2DParaCrystalItem(itfItem, *itf);
        layoutItem->setInterference(itfItem);
    } else if (const auto* itf = dynamic_cast<const InterferenceFinite2DLattice*>(interference)) {
        auto* itfItem = new InterferenceFinite2DLatticeItem();
        setFinite2DLatticeItem(itfItem, *itf);
        layoutItem->setInterference(itfItem);
    } else if (const auto* itf = dynamic_cast<const InterferenceHardDisk*>(interference)) {
        auto* itfItem = new InterferenceHardDiskItem();
        setHardDiskItem(itfItem, *itf);
        layoutItem->setInterference(itfItem);
    } else if (const auto* itf = dynamic_cast<const InterferenceRadialParaCrystal*>(interference)) {
        auto* itfItem = new InterferenceRadialParaCrystalItem();
        setRadialParaCrystalItem(itfItem, *itf);
        layoutItem->setInterference(itfItem);
    } else
        ASSERT(false); // missing implementation for interference type
}

void GUI::Transform::FromDomain::setFormFactor(std::variant<ParticleItem*, MesoCrystalItem*> parent,
                                               const IBornFF* iFormFactor)
{
    if (const auto* formfactor = dynamic_cast<const FormFactorPyramid2*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<Pyramid2Item>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setAlpha(Units::rad2deg(formfactor->alpha()));
    } else if (const auto* formfactor = dynamic_cast<const FormFactorBarGauss*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<BarGaussItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorBarLorentz*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<BarLorentzItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorBox*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<BoxItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorCone*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<ConeItem>(parent);
        formfactorItem->setRadius(formfactor->radius());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setAlpha(Units::rad2deg(formfactor->alpha()));
    } else if (const auto* formfactor = dynamic_cast<const FormFactorPyramid6*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<Pyramid6Item>(parent);
        formfactorItem->setBaseEdge(formfactor->baseEdge());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setAlpha(Units::rad2deg(formfactor->alpha()));
    } else if (const auto* formfactor = dynamic_cast<const FormFactorBipyramid4*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<Bipyramid4Item>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setHeightRatio(formfactor->heightRatio());
        formfactorItem->setAlpha(Units::rad2deg(formfactor->alpha()));
    } else if (const auto* formfactor = dynamic_cast<const FormFactorCylinder*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<CylinderItem>(parent);
        formfactorItem->setRadius(formfactor->radius());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorDodecahedron*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<DodecahedronItem>(parent);
        formfactorItem->setEdge(formfactor->edge());
    } else if (const auto* formfactor =
                   dynamic_cast<const FormFactorEllipsoidalCylinder*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<EllipsoidalCylinderItem>(parent);
        formfactorItem->setRadiusX(formfactor->radiusX());
        formfactorItem->setRadiusY(formfactor->radiusY());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorFullSphere*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<FullSphereItem>(parent);
        formfactorItem->setRadius(formfactor->radius());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorFullSpheroid*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<FullSpheroidItem>(parent);
        formfactorItem->setRadius(formfactor->radius());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorIcosahedron*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<IcosahedronItem>(parent);
        formfactorItem->setEdge(formfactor->edge());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorHemiEllipsoid*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<HemiEllipsoidItem>(parent);
        formfactorItem->setRadiusX(formfactor->radiusX());
        formfactorItem->setRadiusY(formfactor->radiusY());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorPrism3*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<Prism3Item>(parent);
        formfactorItem->setBaseEdge(formfactor->baseEdge());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorPrism6*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<Prism6Item>(parent);
        formfactorItem->setBaseEdge(formfactor->baseEdge());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorPyramid4*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<Pyramid4Item>(parent);
        formfactorItem->setBaseEdge(formfactor->baseEdge());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setAlpha(Units::rad2deg(formfactor->alpha()));
    } else if (const auto* formfactor =
                   dynamic_cast<const FormFactorCosineRippleBox*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<CosineRippleBoxItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor =
                   dynamic_cast<const FormFactorCosineRippleGauss*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<CosineRippleGaussItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor =
                   dynamic_cast<const FormFactorCosineRippleLorentz*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<CosineRippleLorentzItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
    } else if (const auto* formfactor =
                   dynamic_cast<const FormFactorSawtoothRippleBox*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<SawtoothRippleBoxItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setAsymmetry(formfactor->asymmetry());
    } else if (const auto* formfactor =
                   dynamic_cast<const FormFactorSawtoothRippleGauss*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<SawtoothRippleGaussItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setAsymmetry(formfactor->asymmetry());
    } else if (const auto* formfactor =
                   dynamic_cast<const FormFactorSawtoothRippleLorentz*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<SawtoothRippleLorentzItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setWidth(formfactor->width());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setAsymmetry(formfactor->asymmetry());
    } else if (const auto* formfactor = dynamic_cast<const FormFactorPyramid3*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<Pyramid3Item>(parent);
        formfactorItem->setBaseEdge(formfactor->baseEdge());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setAlpha(Units::rad2deg(formfactor->alpha()));
    } else if (const auto* formfactor = dynamic_cast<const FormFactorTruncatedCube*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<TruncatedCubeItem>(parent);
        formfactorItem->setLength(formfactor->length());
        formfactorItem->setRemovedLength(formfactor->removedLength());
    } else if (const auto* formfactor =
                   dynamic_cast<const FormFactorTruncatedSphere*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<TruncatedSphereItem>(parent);
        formfactorItem->setRadius(formfactor->radius());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setRemovedTop(formfactor->removedTop());
    } else if (const auto* formfactor =
                   dynamic_cast<const FormFactorTruncatedSpheroid*>(iFormFactor)) {
        auto* formfactorItem = addFormFactorItem<TruncatedSpheroidItem>(parent);
        formfactorItem->setRadius(formfactor->radius());
        formfactorItem->setHeight(formfactor->height());
        formfactorItem->setHeightFlattening(formfactor->heightFlattening());
        formfactorItem->setRemovedTop(formfactor->removedTop());
    } else
        ASSERT(false);
}
