//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/FromDomain/GUIObjectBuilder.cpp
//! @brief     Implements GUI::Model::ObjectBuilder namespace
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/FromDomain/GUIObjectBuilder.h"
#include "Base/Const/Units.h"
#include "Base/Util/Assert.h"
#include "Core/Simulation/includeSimulations.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Support/Data/SimulationOptionsItem.h"
#include "GUI/View/FromDomain/FromDomain.h"
#include "GUI/View/FromDomain/GUISampleBuilder.h"
#include "Resample/Options/SimulationOptions.h"

namespace {

GISASInstrumentItem* createGISASInstrumentItem(InstrumentItems* model,
                                               const GISASSimulation& simulation,
                                               const QString& name)
{
    auto* result = model->addInstrument<GISASInstrumentItem>();
    result->setInstrumentName(name);
    GUI::Transform::FromDomain::setGISASBeamItem(result->beamItem(), simulation);
    GUI::Transform::FromDomain::setDetector(result, simulation);
    GUI::Transform::FromDomain::setBackground(result, simulation);

    return result;
}

OffSpecularInstrumentItem* createOffSpecularInstrumentItem(InstrumentItems* model,
                                                           const OffSpecularSimulation& simulation,
                                                           const QString& name)
{
    auto* result = model->addInstrument<OffSpecularInstrumentItem>();
    result->setInstrumentName(name);
    GUI::Transform::FromDomain::setOffSpecularBeamItem(result->beamItem(), simulation);
    GUI::Transform::FromDomain::setDetector(result, simulation);
    GUI::Transform::FromDomain::setBackground(result, simulation);

    const double factor = 1. / Units::deg;
    result->alphaAxis().setNbins(simulation.beamAxis()->size());
    result->alphaAxis().setMin(factor * simulation.beamAxis()->lowerBound());
    result->alphaAxis().setMax(factor * simulation.beamAxis()->upperBound());

    return result;
}

SpecularInstrumentItem* createSpecularInstrumentItem(InstrumentItems* model,
                                                     const SpecularSimulation& simulation,
                                                     const QString& name)
{
    auto* result = model->addInstrument<SpecularInstrumentItem>();
    result->setInstrumentName(name);
    GUI::Transform::FromDomain::setSpecularBeamItem(result->beamItem(), simulation);
    GUI::Transform::FromDomain::setBackground(result, simulation);

    return result;
}

} // namespace

//  ************************************************************************************************
//  class implementation
//  ************************************************************************************************

void GUI::Transform::FromDomain::populateInstrumentItems(InstrumentItems* instrumentItems,
                                                         const ISimulation& simulation,
                                                         const QString& instrument_name)
{
    ASSERT(instrumentItems);

    QString name = instrument_name.isEmpty()
                       ? QString::fromStdString(simulation.instrument().className())
                       : instrument_name;

    if (const auto* gisasSimulation = dynamic_cast<const GISASSimulation*>(&simulation))
        createGISASInstrumentItem(instrumentItems, *gisasSimulation, name);
    else if (const auto* offSpecSimulation =
                 dynamic_cast<const OffSpecularSimulation*>(&simulation))
        createOffSpecularInstrumentItem(instrumentItems, *offSpecSimulation, name);
    else if (const auto* spec_simulation = dynamic_cast<const SpecularSimulation*>(&simulation))
        createSpecularInstrumentItem(instrumentItems, *spec_simulation, name);
    else
        ASSERT(0);
}

void GUI::Transform::FromDomain::populateSimulationOptions(SimulationOptionsItem* item,
                                                           const ISimulation& simulation)
{
    if (simulation.options().isIntegrate())
        item->setUseMonteCarloIntegration(
            static_cast<unsigned>(simulation.options().getMcPoints()));
    else
        item->setUseAnalytical();

    item->setUseAverageMaterials(simulation.options().useAvgMaterials());
    item->setIncludeSpecularPeak(simulation.options().includeSpecular());
}

MultiLayerItem* GUI::Transform::FromDomain::createMultiLayerItem(const MultiLayer& node,
                                                                 const QString& node_name /*= ""*/)
{
    return GUISampleBuilder().createMultiLayerItem(node, node_name);
}
