//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/FromDomain/GUISampleBuilder.cpp
//! @brief     Implements class GUISampleBuilder
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/FromDomain/GUISampleBuilder.h"
#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/MaterialItem.h"
#include "GUI/Model/Sample/MaterialItems.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/View/FromDomain/FromDomain.h"
#include "Sample/Aggregate/ParticleLayout.h"
#include "Sample/Multilayer/Layer.h"
#include "Sample/Multilayer/MultiLayer.h"
#include "Sample/Particle/Crystal.h"
#include "Sample/Particle/MesoCrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleCoreShell.h"

namespace GUI::Transform::FromDomain {

MultiLayerItem* GUISampleBuilder::createMultiLayerItem(const MultiLayer& multiLayer,
                                                       const QString& nodeName)
{
    m_multilayerItem = new MultiLayerItem();
    m_multilayerItem->setSampleName(
        nodeName.isEmpty() ? QString::fromStdString(multiLayer.sampleName()) : nodeName);
    m_multilayerItem->crossCorrLength().set(multiLayer.crossCorrLength());
    m_multilayerItem->setExternalField(multiLayer.externalField());

    //  iterate over layers
    for (size_t layerIndex = 0; layerIndex < multiLayer.numberOfLayers(); layerIndex++) {
        const auto* layer = multiLayer.layer(layerIndex);

        const LayerInterface* top_interface =
            layerIndex == 0 ? nullptr : multiLayer.layerInterface(layerIndex - 1);

        auto* layerItem = m_multilayerItem->addLayer();
        layerItem->setMaterial(createMaterialFromDomain(layer->material()));
        layerItem->setIsTopLayer(layerIndex == 0);
        layerItem->setIsBottomLayer(layerIndex == multiLayer.numberOfLayers() - 1);

        FromDomain::setLayerItem(layerItem, layer, top_interface);

        // iterate over particle layouts
        for (const auto* layout : layer->layouts()) {
            auto* layoutItem = layerItem->addLayout();
            layoutItem->ownDensity().set(layout->totalParticleSurfaceDensity());
            layoutItem->weight().set(layout->weight());
            FromDomain::setInterference(layoutItem, layout->interferenceFunction());

            // create particles/particle compositions
            for (const auto& particle : layout->particles())
                copyParticle(particle.get(),
                             [=](ItemWithParticles* p) { layoutItem->addParticle(p); });
        }
    }

    return m_multilayerItem;
}

MaterialItem* GUISampleBuilder::createMaterialFromDomain(const Material* material)
{
    QString materialName = QString::fromStdString(material->materialName());

    if (auto* materialItem = m_multilayerItem->materialItems().materialFromName(materialName))
        return materialItem;

    complex_t material_data = material->materialData();
    MaterialItem* materialItem(nullptr);
    if (material->typeID() == MATERIAL_TYPES::RefractiveMaterial) {
        materialItem = m_multilayerItem->materialItems().addRefractiveMaterial(
            materialName, material_data.real(), material_data.imag());
    } else if (material->typeID() == MATERIAL_TYPES::MaterialBySLD) {
        materialItem = m_multilayerItem->materialItems().addSLDMaterial(
            materialName, material_data.real(), material_data.imag());
    } else
        ASSERT(0);

    materialItem->setMagnetization(material->magnetization());
    return materialItem;
}

void GUISampleBuilder::copyParticle(ParticleItem* particleItem, const Particle* particle)
{
    particleItem->setAbundance(particle->abundance());
    particleItem->setPosition(particle->particlePosition());
    particleItem->setMaterial(createMaterialFromDomain(particle->material()));
    FromDomain::setRotation(particleItem, particle->rotation());
    FromDomain::setFormFactor(particleItem, particle->formfactor_at_bottom());
}

void GUISampleBuilder::copyParticle(const IParticle* iparticle,
                                    std::function<void(ItemWithParticles*)> addToParent)
{
    if (const auto* particle = dynamic_cast<const Particle*>(iparticle)) {
        auto* particle_item = new ParticleItem(&m_multilayerItem->materialItems());
        copyParticle(particle_item, particle);
        addToParent(particle_item);
    } else if (const auto* coreshell = dynamic_cast<const ParticleCoreShell*>(iparticle)) {
        if (!coreshell->coreParticle())
            throw Exception("Core/shell particle has no core particle.");
        if (!coreshell->shellParticle())
            throw Exception("Core/shell particle has no shell particle.");

        auto* coreShellItem = new ParticleCoreShellItem(&m_multilayerItem->materialItems());
        coreShellItem->setAbundance(coreshell->abundance());
        coreShellItem->setPosition(coreshell->particlePosition());
        FromDomain::setRotation(coreShellItem, coreshell->rotation());

        if (coreshell->coreParticle()) {
            coreShellItem->createCore(&m_multilayerItem->materialItems());
            copyParticle(coreShellItem->core(), coreshell->coreParticle());
        }
        if (coreshell->shellParticle()) {
            coreShellItem->createShell(&m_multilayerItem->materialItems());
            copyParticle(coreShellItem->shell(), coreshell->shellParticle());
        }

        addToParent(coreShellItem);
    } else if (const auto* meso = dynamic_cast<const MesoCrystal*>(iparticle)) {
        if (!meso->particleStructure().basis())
            throw Exception("Meso crystal has no basis.");

        auto* mesoItem = new MesoCrystalItem(&m_multilayerItem->materialItems());
        mesoItem->setAbundance(meso->abundance());
        mesoItem->setPosition(meso->particlePosition());
        FromDomain::setFormFactor(mesoItem, meso->outerShape());
        FromDomain::setRotation(mesoItem, meso->rotation());

        const Lattice3D* lattice = meso->particleStructure().lattice();
        mesoItem->setVectorA(lattice->basisVectorA());
        mesoItem->setVectorB(lattice->basisVectorB());
        mesoItem->setVectorC(lattice->basisVectorC());
        addToParent(mesoItem);

        copyParticle(meso->particleStructure().basis(),
                     [=](ItemWithParticles* p) { mesoItem->setBasis(p); });

    } else if (const auto* particleComposition =
                   dynamic_cast<const ParticleComposition*>(iparticle)) {
        auto* particleCompositionItem =
            new ParticleCompositionItem(&m_multilayerItem->materialItems());

        particleCompositionItem->setAbundance(particleComposition->abundance());
        particleCompositionItem->setPosition(particleComposition->particlePosition());
        FromDomain::setRotation(particleCompositionItem, particleComposition->rotation());
        addToParent(particleCompositionItem);

        for (const auto* p : particleComposition->particles())
            copyParticle(p, [=](ItemWithParticles* p) { particleCompositionItem->addParticle(p); });
    }
}

} // namespace GUI::Transform::FromDomain
