//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/DepthProbeInstrumentEditor.cpp
//! @brief     Implements class DepthProbeInstrumentEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/DepthProbeInstrumentEditor.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/View/Instrument/AxisPropertyEditor.h"
#include "GUI/View/Instrument/DistributionEditor.h"
#include "GUI/View/Instrument/InclinationAnglesEditor.h"
#include "GUI/View/Instrument/PolarizationAnalysisEditor.h"
#include "GUI/View/Instrument/SphericalAxisEditor.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include <QGridLayout>
#include <QVBoxLayout>

DepthProbeInstrumentEditor::DepthProbeInstrumentEditor(QWidget* parent,
                                                       DepthProbeInstrumentItem* instrument)
    : QWidget(parent)
{
    ASSERT(instrument);
    auto* layout = new QVBoxLayout(this);
    layout->setContentsMargins(0, 0, 0, 0);

    auto* parametersGroupBox = new QGroupBox(this);
    parametersGroupBox->setTitle("Parameters");
    auto* grid = new QGridLayout(parametersGroupBox);
    grid->setColumnStretch(0, 1);
    grid->setColumnStretch(1, 1);
    grid->setColumnStretch(2, 1);

    // valid input range is ]0,inf[
    MeanConfig wave_cfg{boost::numeric::interval<double>(std::numeric_limits<double>::min(),
                                                         std::numeric_limits<double>::max()),
                        4, true};
    auto* m_wavelengthEditor =
        new DistributionEditor("Wavelength [nm]", wave_cfg, GUI::ID::Distributions::Symmetric, this,
                               instrument->beamItem()->wavelengthItem());
    grid->addWidget(m_wavelengthEditor, 1, 0);

    auto* inclinationEditor =
        new InclinationAnglesEditor(this, instrument->beamItem()->inclinationAngleItem());
    grid->addWidget(inclinationEditor, 1, 1);

    auto* depthAxisEditor = new AxisPropertyEditor(this, "Depth axis", &instrument->zAxis());
    grid->addWidget(depthAxisEditor, 1, 2);

    GroupBoxCollapser::installIntoGroupBox(parametersGroupBox);
    layout->addWidget(parametersGroupBox);

    auto* polpairAnalysisEditor = new PolarizationAnalysisEditor(this, instrument);
    layout->addWidget(polpairAnalysisEditor);

    layout->addStretch();

    connect(m_wavelengthEditor, &DistributionEditor::distributionChanged, this,
            &DepthProbeInstrumentEditor::dataChanged);
    connect(inclinationEditor, &InclinationAnglesEditor::dataChanged, this,
            &DepthProbeInstrumentEditor::dataChanged);
    connect(depthAxisEditor, &AxisPropertyEditor::dataChanged, this,
            &DepthProbeInstrumentEditor::dataChanged);
    connect(polpairAnalysisEditor, &PolarizationAnalysisEditor::dataChanged, this,
            &DepthProbeInstrumentEditor::dataChanged);
}
