//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/DistributionEditor.cpp
//! @brief     Implements classes DistributionSelector and DistributionEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/DistributionEditor.h"
#include "GUI/Model/CatDevice/DistributionItemCatalog.h"
#include "GUI/Model/Descriptor/DistributionItems.h"
#include "GUI/Model/Device/BeamDistributionItem.h"
#include "GUI/Support/XML/Backup.h"
#include "GUI/Support/XML/Streamer.h"
#include "GUI/View/Instrument/DistributionForms.h"
#include "GUI/View/Instrument/EditDistributionDialog.h"
#include "GUI/View/PropertyEditor/GroupInfoBox.h"
#include "GUI/View/Tool/WidgetUtils.h"
#include <QComboBox>
#include <QFormLayout>
#include <QStackedLayout>
#include <utility>

//==================================================================================================
// DistributionSelector
//==================================================================================================

DistributionSelector::DistributionSelector(std::optional<MeanConfig> mean_config,
                                           GUI::ID::Distributions distributions, QWidget* parent,
                                           BeamDistributionItem* item)
    : QWidget(parent)
    , m_item(item)
    , m_meanConfig(std::move(mean_config))
    , m_distributions(distributions)
{
    ASSERT(item);
    m_formLayout = new QFormLayout(this);
    m_formLayout->setMargin(0);

    m_distributionCombo =
        GUI::Util::createSelectionCombo(this, item->distributionSelection(), [=](int) {
            createDistributionWidgets();
            emit distributionChanged();
        });
    m_formLayout->addRow("Distribution:", m_distributionCombo);

    createDistributionWidgets();
}

void DistributionSelector::createDistributionWidgets()
{
    while (m_formLayout->rowCount() > 1)
        m_formLayout->removeRow(1);

    DistributionForm* form = nullptr;
    if (m_item->distribution()->is<DistributionCosineItem>())
        form = new CosineDistributionForm(m_meanConfig, this);
    else if (m_item->distribution()->is<DistributionGateItem>())
        form = new GateDistributionForm(this);
    else if (m_item->distribution()->is<DistributionGaussianItem>())
        form = new GaussDistributionForm(m_meanConfig, this);
    else if (m_item->distribution()->is<DistributionLogNormalItem>())
        form = new LogNormalDistributionForm(this);
    else if (m_item->distribution()->is<DistributionLorentzItem>())
        form = new LorentzDistributionForm(m_meanConfig, this);
    else if (m_item->distribution()->is<DistributionNoneItem>())
        form = new NoneDistributionForm(m_meanConfig, this);
    else if (m_item->distribution()->is<DistributionTrapezoidItem>())
        form = new TrapezoidDistributionForm(this);

    ASSERT(form);
    form->setupDistribution(m_item);
    m_formLayout->addRow(form);
    connect(form, &CosineDistributionForm::distributionChanged, this,
            &DistributionSelector::distributionChanged);
}

BeamDistributionItem* DistributionSelector::item() const
{
    return m_item;
}

const std::optional<MeanConfig>& DistributionSelector::meanConfig() const
{
    return m_meanConfig;
}

GUI::ID::Distributions DistributionSelector::distributions() const
{
    return m_distributions;
}

void DistributionSelector::refresh()
{
    QSignalBlocker b(m_distributionCombo);
    m_distributionCombo->setCurrentIndex(m_item->distributionSelection().currentIndex());
    createDistributionWidgets();
}


//==================================================================================================
// DistributionEditor
//==================================================================================================

DistributionEditor::DistributionEditor(const QString& title,
                                       const std::optional<MeanConfig>& mean_config,
                                       GUI::ID::Distributions distributions, QWidget* parent,
                                       BeamDistributionItem* item)
    : QWidget(parent)
    , m_title(title)
{
    auto* layout = new QVBoxLayout(this);

    auto* gbox = new GroupInfoBox(title, this);
    auto* glayout = new QVBoxLayout(gbox);
    m_selector = new DistributionSelector(mean_config, distributions, gbox, item);
    glayout->addWidget(m_selector);
    layout->addWidget(gbox);

    connect(m_selector, &DistributionSelector::distributionChanged, this,
            &DistributionEditor::distributionChanged);
    connect(gbox, &GroupInfoBox::clicked, this, &DistributionEditor::showDialog);
}

void DistributionEditor::updateData()
{
    m_selector->refresh();
}

void DistributionEditor::showDialog()
{
    showEditDistributionDialog(m_selector, m_title);
}

//--------------------------------------------------------------------------------------------------
// free functions
//--------------------------------------------------------------------------------------------------

void showEditDistributionDialog(DistributionSelector* selector, const QString& title)
{
    BeamDistributionItem* item = selector->item();
    if (!item)
        return;

    const QByteArray backup = GUI::Util::createBackup(item);
    bool modifiedInDialog = false;

    EditDistributionDialog dialog(item, title, selector->meanConfig(), selector->distributions(),
                                  selector);

    QObject::connect(&dialog, &EditDistributionDialog::distributionChanged,
                     [&] { modifiedInDialog = true; });
    if (dialog.exec() != QDialog::Accepted)
        GUI::Util::restoreBackup(item, backup);
    else {
        if (modifiedInDialog) {
            selector->refresh();
            selector->distributionChanged();
        }
    }
}
