//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/GISASBeamEditor.cpp
//! @brief     Implements class GISASBeamEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/GISASBeamEditor.h"
#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include "GUI/Model/Device/BeamAngleItems.h"
#include "GUI/Model/Device/BeamItems.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/View/Instrument/DistributionEditor.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include <QDoubleValidator>
#include <QFormLayout>
#include <QGridLayout>
#include <QLineEdit>

GISASBeamEditor::GISASBeamEditor(QWidget* parent, BeamItem* item)
    : QGroupBox("Beam parameters", parent)
{
    ASSERT(item);
    auto* grid = new QGridLayout(this);
    grid->setColumnStretch(0, 1);
    grid->setColumnStretch(1, 1);
    grid->setColumnStretch(2, 1);

    auto* form = new QFormLayout();
    grid->addLayout(form, 0, 0);
    auto* intensityEditor = new QLineEdit(this);
    intensityEditor->setToolTip("Beam intensity in neutrons (or gammas per second.");
    auto* validator = new QDoubleValidator(intensityEditor);
    validator->setNotation(QDoubleValidator::ScientificNotation);
    validator->setRange(0.0, 1e+32, 1000);
    intensityEditor->setValidator(validator);
    form->addRow("Intensity:", intensityEditor);

    // valid input range is ]0,inf[
    MeanConfig wave_cfg{boost::numeric::interval<double>(std::numeric_limits<double>::min(),
                                                         std::numeric_limits<double>::max()),
                        4, true};
    auto* wavelengthEditor = new DistributionEditor(
        "Wavelength [nm]", wave_cfg, GUI::ID::Distributions::All, this, item->wavelengthItem());
    grid->addWidget(wavelengthEditor, 1, 0);

    // valid input range is [0,90]
    MeanConfig incl_cfg{boost::numeric::interval<double>(0.0, 90.0), 3, false};
    auto* inclinationEditor =
        new DistributionEditor("Inclination angle [deg]", incl_cfg, GUI::ID::Distributions::All,
                               this, item->inclinationAngleItem());
    grid->addWidget(inclinationEditor, 1, 1);

    // valid input range is [-90,90]
    MeanConfig azim_cfg{boost::numeric::interval<double>(-90.0, 90.0), 3, false};
    auto* azimuthalEditor =
        new DistributionEditor("Azimuthal angle [deg]", azim_cfg, GUI::ID::Distributions::All, this,
                               item->azimuthalAngleItem());
    grid->addWidget(azimuthalEditor, 1, 2);

    intensityEditor->setText(QString::number(item->intensity()));

    GroupBoxCollapser::installIntoGroupBox(this);

    connect(wavelengthEditor, &DistributionEditor::distributionChanged, this,
            &GISASBeamEditor::dataChanged);
    connect(inclinationEditor, &DistributionEditor::distributionChanged, this,
            &GISASBeamEditor::dataChanged);
    connect(azimuthalEditor, &DistributionEditor::distributionChanged, this,
            &GISASBeamEditor::dataChanged);

    connect(intensityEditor, &QLineEdit::editingFinished, [=]() {
        item->setIntensity(intensityEditor->text().toDouble());
        emit dataChanged();
    });
}
