//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InclinationAnglesEditor.cpp
//! @brief     Defines class InclinationAnglesEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InclinationAnglesEditor.h"
#include "GUI/Model/BaseItem/GroupItem.h"
#include "GUI/Model/Device/PointwiseAxisItem.h"
#include "GUI/Model/Device/SpecularBeamInclinationItem.h"
#include "GUI/View/Instrument/DistributionEditor.h"
#include "GUI/View/Instrument/SphericalAxisEditor.h"
#include "GUI/View/PropertyEditor/GroupInfoBox.h"

#include <QComboBox>
#include <QFormLayout>

namespace {
static constexpr auto title = "Inclination angles [deg]";
}

InclinationAnglesEditor::InclinationAnglesEditor(QWidget* parent, SpecularBeamInclinationItem* item)
    : QWidget(parent)
    , m_item(item)
{
    auto* layout = new QVBoxLayout(this);

    auto* gbox = new GroupInfoBox(title, this);
    auto* gform = new QFormLayout(gbox);
    connect(gbox, &GroupInfoBox::clicked, this, &InclinationAnglesEditor::showDialog);
    layout->addWidget(gbox);

    m_distributionEditor =
        new DistributionSelector(std::nullopt, GUI::ID::Distributions::Symmetric, gbox, m_item);
    gform->addRow(m_distributionEditor);

    m_typeComboBox = new QComboBox(gbox);
    m_typeComboBox->addItem("Uniform axis");
    m_typeComboBox->addItem("Non-uniform axis");
    gform->addRow("Axis type:", m_typeComboBox);
    connect(m_typeComboBox, qOverload<int>(&QComboBox::currentIndexChanged), this,
            &InclinationAnglesEditor::onAxisTypeSelected);

    m_form = new SphericalAxisForm(gform, gbox);

    const int idx = m_item->pointwiseAlphaAxisSelected() ? 1 : 0;
    ASSERT(idx != -1);
    if (idx != m_typeComboBox->currentIndex())
        m_typeComboBox->setCurrentIndex(idx);
    else
        onAxisTypeSelected(idx); // trigger update manually
    m_typeComboBox->setEnabled(m_item->pointwiseAlphaAxisDefined());

    connect(m_distributionEditor, &DistributionSelector::distributionChanged, this,
            &InclinationAnglesEditor::dataChanged);
    connect(m_form, &SphericalAxisForm::dataChanged, this, &InclinationAnglesEditor::dataChanged);
}

void InclinationAnglesEditor::updateIndicators()
{
    m_form->updateData();
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void InclinationAnglesEditor::onAxisTypeSelected(int index)
{
    if (m_item) {
        if (index == 0 && m_item->pointwiseAlphaAxisSelected()) {
            m_item->selectUniformAxis();
            emit dataChanged();
        } else if (index == 1 && !m_item->pointwiseAlphaAxisSelected()) {
            m_item->selectPointwiseAxis();
            emit dataChanged();
        }

        m_form->setItem(m_item->alphaAxis());
        m_form->setEnabled(index == 0);
    }
}

void InclinationAnglesEditor::showDialog()
{
    showEditDistributionDialog(m_distributionEditor, title);
}
