//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/OffSpecularBeamEditor.cpp
//! @brief     Implements class OffSpecularBeamEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/OffSpecularBeamEditor.h"
#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include "GUI/Model/Device/BeamAngleItems.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/View/Instrument/AxisPropertyEditor.h"
#include "GUI/View/Instrument/DistributionEditor.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include <QDoubleValidator>
#include <QFormLayout>
#include <QGridLayout>
#include <QLineEdit>

OffSpecularBeamEditor::OffSpecularBeamEditor(QWidget* parent, OffSpecularInstrumentItem* item)
    : QGroupBox("Beam parameters", parent)
{
    ASSERT(item);

    auto* grid = new QGridLayout(this);
    grid->setColumnStretch(0, 1);
    grid->setColumnStretch(1, 1);
    grid->setColumnStretch(2, 1);

    auto* form = new QFormLayout();
    grid->addLayout(form, 0, 0);
    auto* intensityEditor = new QLineEdit(this);
    intensityEditor->setToolTip("Beam intensity in neutrons (or gammas per second.");
    auto* validator = new QDoubleValidator(intensityEditor);
    validator->setNotation(QDoubleValidator::ScientificNotation);
    validator->setRange(0.0, 1e+32, 1000);
    intensityEditor->setValidator(validator);
    form->addRow("Intensity:", intensityEditor);

    // valid input range is ]0,inf[
    MeanConfig wave_cfg{boost::numeric::interval<double>(std::numeric_limits<double>::min(),
                                                         std::numeric_limits<double>::max()),
                        4, true};
    auto* wavelengthEditor =
        new DistributionEditor("Wavelength [nm]", wave_cfg, GUI::ID::Distributions::All, this,
                               item->beamItem()->wavelengthItem());
    grid->addWidget(wavelengthEditor, 1, 0);

    auto* inclinationEditor =
        new AxisPropertyEditor(this, "Inclination angle [deg]", &item->alphaAxis());
    grid->addWidget(inclinationEditor, 1, 1);

    // valid input range is [-90,90]
    MeanConfig azim_cfg{boost::numeric::interval<double>(-90.0, 90.0), 3, false};
    auto* azimuthalEditor =
        new DistributionEditor("Azimuthal angle [deg]", azim_cfg, GUI::ID::Distributions::All, this,
                               item->beamItem()->azimuthalAngleItem());
    grid->addWidget(azimuthalEditor, 1, 2);

    intensityEditor->setText(QString::number(item->beamItem()->intensity()));

    GroupBoxCollapser::installIntoGroupBox(this);

    connect(wavelengthEditor, &DistributionEditor::distributionChanged, this,
            &OffSpecularBeamEditor::dataChanged);
    connect(inclinationEditor, &AxisPropertyEditor::dataChanged, this,
            &OffSpecularBeamEditor::dataChanged);
    connect(azimuthalEditor, &DistributionEditor::distributionChanged, this,
            &OffSpecularBeamEditor::dataChanged);
    connect(intensityEditor, &QLineEdit::editingFinished, [=]() {
        const double newValue = intensityEditor->text().toDouble();
        if (item->beamItem()->intensity() != newValue) {
            item->beamItem()->setIntensity(newValue);
            emit dataChanged();
        }
    });
}
