//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/PolarizationAnalysisEditor.cpp
//! @brief     Implements class PolarizationAnalysisEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/PolarizationAnalysisEditor.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/View/Edit/DoubleSpinBox.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include "GUI/View/Tool/WidgetUtils.h"
#include <QCheckBox>
#include <QFormLayout>
#include <QLabel>

PolarizationAnalysisEditor::PolarizationAnalysisEditor(QWidget* parent, InstrumentItem* instrument)
    : QGroupBox("Polarization analysis", parent)
    , m_instrument(instrument)
{
    ASSERT(instrument);

    auto* layout = new QVBoxLayout(this);

    auto* checkbox = new QCheckBox(QString("Enable %1").arg(title()));
    checkbox->setChecked(m_instrument->withPolarizerAnalyzer());
    layout->addWidget(checkbox);

    auto* polarizerAnalyzerWidget = new QWidget(this);
    auto* formlayout = new QFormLayout(polarizerAnalyzerWidget);
    formlayout->setFieldGrowthPolicy(QFormLayout::FieldsStayAtSizeHint);

    addVector(formlayout, m_instrument->polarization());
    addVector(formlayout, m_instrument->analyzerDirection());
    formlayout->addRow(GUI::Util::labelWithUnit(m_instrument->analyzerEfficiency()) + ":",
                       createSpinBox(polarizerAnalyzerWidget, m_instrument->analyzerEfficiency()));
    formlayout->addRow(
        GUI::Util::labelWithUnit(m_instrument->analyzerTotalTransmission()) + ":",
        createSpinBox(polarizerAnalyzerWidget, m_instrument->analyzerTotalTransmission()));

    layout->addWidget(polarizerAnalyzerWidget);
    polarizerAnalyzerWidget->setVisible(m_instrument->withPolarizerAnalyzer());

    connect(checkbox, &QCheckBox::toggled, [=](bool b) {
        polarizerAnalyzerWidget->setVisible(b);
        m_instrument->setWithPolarizerAnalyzer(b);
        emit dataChanged();
    });

    GroupBoxCollapser::installIntoGroupBox(this);
}

DoubleSpinBox* PolarizationAnalysisEditor::createSpinBox(QWidget* parent, const DoubleDescriptor& d)
{
    auto* sb = new DoubleSpinBox(parent, d);
    connect(sb, &DoubleSpinBox::baseValueChanged, [=](double v) {
        if (d.get() != v) {
            d.set(v);
            emit dataChanged();
        }
    });
    return sb;
}

void PolarizationAnalysisEditor::addVector(QFormLayout* parentLayout, const VectorDescriptor& d)
{
    auto* layout = new QHBoxLayout;

    const auto add = [&](const DoubleDescriptor& d) {
        layout->addWidget(new QLabel(GUI::Util::labelWithUnit(d) + ":"));
        layout->addWidget(createSpinBox(parentLayout->parentWidget(), d));
    };

    add(d.x);
    add(d.y);
    add(d.z);

    layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Expanding));

    parentLayout->addRow(d.label + ":", layout);
}
