//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/SpecularBeamEditor.cpp
//! @brief     Defines class SpecularBeamEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/SpecularBeamEditor.h"
#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include "GUI/Model/Device/BeamItems.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/View/Instrument/DistributionEditor.h"
#include "GUI/View/Instrument/FootprintCorrectionEditor.h"
#include "GUI/View/Instrument/InclinationAnglesEditor.h"
#include "GUI/View/Instrument/InstrumentEditController.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include <QDoubleValidator>
#include <QFormLayout>
#include <QGridLayout>
#include <QLineEdit>

SpecularBeamEditor::SpecularBeamEditor(QWidget* parent, SpecularBeamItem* item,
                                       InstrumentEditController* ec)
    : QGroupBox("Beam parameters", parent)
{
    ASSERT(item);

    auto* grid = new QGridLayout(this);
    grid->setColumnStretch(0, 1);
    grid->setColumnStretch(1, 1);
    grid->setColumnStretch(2, 1);

    auto* form = new QFormLayout();
    grid->addLayout(form, 0, 0);
    auto* intensityLineEdit = new QLineEdit(this);
    intensityLineEdit->setToolTip("Beam intensity in neutrons (or gammas per second.");
    auto* validator = new QDoubleValidator(intensityLineEdit);
    validator->setNotation(QDoubleValidator::ScientificNotation);
    validator->setRange(0.0, 1e+32, 1000);
    intensityLineEdit->setValidator(validator);
    form->addRow("Intensity:", intensityLineEdit);

    // valid input range is ]0,inf[
    MeanConfig wave_cfg{boost::numeric::interval<double>(std::numeric_limits<double>::min(),
                                                         std::numeric_limits<double>::max()),
                        4, true};
    auto* wavelengthEditor =
        new DistributionEditor("Wavelength [nm]", wave_cfg, GUI::ID::Distributions::Symmetric, this,
                               item->wavelengthItem());
    grid->addWidget(wavelengthEditor, 1, 0);

    auto* inclinationEditor = new InclinationAnglesEditor(this, item->inclinationAngleItem());
    grid->addWidget(inclinationEditor, 1, 1);

    auto* footprintEditor = new FootprintCorrectionEditor(this, item);
    grid->addWidget(footprintEditor, 1, 2);

    intensityLineEdit->setText(QString::number(item->intensity()));

    GroupBoxCollapser::installIntoGroupBox(this);

    // order of the next two connections is important! Indicators have to be recalculated first,
    // then updated (recalculation is done in EditController)
    connect(wavelengthEditor, &DistributionEditor::distributionChanged, ec,
            &InstrumentEditController::notifyWavelengthDistributionChanged);
    connect(wavelengthEditor, &DistributionEditor::distributionChanged, inclinationEditor,
            &InclinationAnglesEditor::updateIndicators);

    // order of the next two connections is important! Wavelength has to be recalculated first,
    // then updated (recalculation is done in EditController)
    connect(inclinationEditor, &InclinationAnglesEditor::dataChanged, ec,
            &InstrumentEditController::notifyInclinationAxisChanged);
    connect(inclinationEditor, &InclinationAnglesEditor::dataChanged, wavelengthEditor,
            &DistributionEditor::updateData);

    connect(footprintEditor, &FootprintCorrectionEditor::dataChanged, this,
            &SpecularBeamEditor::dataChanged);
    connect(intensityLineEdit, &QLineEdit::editingFinished, [=]() {
        const double newValue = intensityLineEdit->text().toDouble();
        if (item->intensity() != newValue) {
            item->setIntensity(newValue);
            emit dataChanged();
        }
    });
}
