//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Particle/ParticleComposition.cpp
//! @brief     Implements class ParticleComposition.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Particle/ParticleComposition.h"
#include "Base/Util/Assert.h"
#include "Sample/Scattering/Rotations.h"

ParticleComposition::ParticleComposition() {}

ParticleComposition::~ParticleComposition() = default;

ParticleComposition* ParticleComposition::clone() const
{
    auto* result = new ParticleComposition();
    result->setAbundance(m_abundance);
    for (size_t index = 0; index < m_particles.size(); ++index)
        result->addParticle(*m_particles[index]);
    if (m_rotation)
        result->setRotation(*m_rotation);
    result->setParticlePosition(m_position);
    return result;
}

std::vector<const INode*> ParticleComposition::nodeChildren() const
{
    std::vector<const INode*> result = IParticle::nodeChildren();
    for (const auto& particle : m_particles)
        result.emplace_back(particle.get());
    return result;
}

#ifndef SWIG
std::vector<std::unique_ptr<IParticle>> ParticleComposition::decompose() const
{
    std::vector<std::unique_ptr<IParticle>> result;
    const auto* rot = rotation();
    const auto& translation = particlePosition();
    for (const auto& particle : m_particles) {
        const auto sublist = particle->decompose();
        for (const auto& subparticle : sublist) {
            IParticle* new_subparticle = subparticle->clone();
            if (rot)
                new_subparticle->rotate(*rot);
            new_subparticle->translate(translation);
            result.emplace_back(new_subparticle);
        }
    }
    return result;
}
#endif

void ParticleComposition::addParticle(const IParticle& particle)
{
    m_particles.emplace_back(particle.clone());
}

void ParticleComposition::addParticle(const IParticle& particle, R3 position)
{
    m_particles.emplace_back(particle.clone()->translate(position));
}

// Please note, that positions is not const reference here. This is intentional, to
// enable python lists to std::vector conversion
void ParticleComposition::addParticles(const IParticle& particle, std::vector<R3> positions)
{
    for (size_t i = 0; i < positions.size(); ++i)
        addParticle(particle, positions[i]);
}

std::vector<const IParticle*> ParticleComposition::particles() const
{
    std::vector<const IParticle*> result;
    for (const auto& P_particle : m_particles)
        result.push_back(P_particle.get());
    return result;
}
