#include "Base/Const/Units.h"
#include "Base/Py/PyFmt.h"
#include "Core/Export/PyFmt2.h"
#include "Core/Export/PyFmtLimits.h"
#include "Param/Distrib/Distributions.h"
#include "Param/Distrib/ParameterDistribution.h"
#include "Sample/Correlations/Profiles1D.h"
#include "Sample/Correlations/Profiles2D.h"
#include "Sample/HardParticle/FormFactorBar.h"
#include "Sample/HardParticle/FormFactorBipyramid4.h"
#include "Sample/HardParticle/FormFactorBox.h"
#include "Sample/HardParticle/FormFactorCantellatedCube.h"
#include "Sample/HardParticle/FormFactorCone.h"
#include "Sample/HardParticle/FormFactorCosineRipple.h"
#include "Sample/HardParticle/FormFactorCylinder.h"
#include "Sample/HardParticle/FormFactorDodecahedron.h"
#include "Sample/HardParticle/FormFactorEllipsoidalCylinder.h"
#include "Sample/HardParticle/FormFactorFullSphere.h"
#include "Sample/HardParticle/FormFactorFullSpheroid.h"
#include "Sample/HardParticle/FormFactorHemiEllipsoid.h"
#include "Sample/HardParticle/FormFactorHollowSphere.h"
#include "Sample/HardParticle/FormFactorIcosahedron.h"
#include "Sample/HardParticle/FormFactorLongBoxGauss.h"
#include "Sample/HardParticle/FormFactorLongBoxLorentz.h"
#include "Sample/HardParticle/FormFactorPrism3.h"
#include "Sample/HardParticle/FormFactorPrism6.h"
#include "Sample/HardParticle/FormFactorPyramid2.h"
#include "Sample/HardParticle/FormFactorPyramid3.h"
#include "Sample/HardParticle/FormFactorPyramid4.h"
#include "Sample/HardParticle/FormFactorPyramid6.h"
#include "Sample/HardParticle/FormFactorSawtoothRipple.h"
#include "Sample/HardParticle/FormFactorTruncatedCube.h"
#include "Sample/HardParticle/FormFactorTruncatedSphere.h"
#include "Sample/HardParticle/FormFactorTruncatedSpheroid.h"
#include "Sample/Interface/LayerRoughness.h"
#include "Sample/SoftParticle/FormFactorGauss.h"
#include "Sample/SoftParticle/FormFactorSphereGaussianRadius.h"
#include "Sample/SoftParticle/FormFactorSphereLogNormalRadius.h"
#include "Tests/GTestWrapper/google_test.h"

class PythonFormattingTest : public ::testing::Test {
};

TEST_F(PythonFormattingTest, ValueTimesUnits)
{
    EXPECT_EQ("2*nm", Py::Fmt::printValue(2.0, "nm"));
    EXPECT_EQ("2*deg", Py::Fmt::printValue(2.0 * Units::deg, "rad"));
    EXPECT_EQ("2.0", Py::Fmt::printValue(2.0, ""));
}

TEST_F(PythonFormattingTest, RealLimits)
{
    EXPECT_EQ("RealLimits.lowerLimited(1.0)",
              Py::Fmt::printRealLimits(RealLimits::lowerLimited(1.0)));
    EXPECT_EQ("RealLimits.lowerLimited(1*nm)",
              Py::Fmt::printRealLimits(RealLimits::lowerLimited(1.0), "nm"));
    EXPECT_EQ("RealLimits.lowerLimited(1*deg)",
              Py::Fmt::printRealLimits(RealLimits::lowerLimited(1.0 * Units::deg), "rad"));

    EXPECT_EQ("RealLimits.upperLimited(1.0)",
              Py::Fmt::printRealLimits(RealLimits::upperLimited(1.0)));
    EXPECT_EQ("RealLimits.upperLimited(1*nm)",
              Py::Fmt::printRealLimits(RealLimits::upperLimited(1.0), "nm"));
    EXPECT_EQ("RealLimits.upperLimited(1*deg)",
              Py::Fmt::printRealLimits(RealLimits::upperLimited(1.0 * Units::deg), "rad"));

    EXPECT_EQ("RealLimits.limited(1.0, 2.0)",
              Py::Fmt::printRealLimits(RealLimits::limited(1.0, 2.0)));
    EXPECT_EQ("RealLimits.limited(1*nm, 2*nm)",
              Py::Fmt::printRealLimits(RealLimits::limited(1.0, 2.0), "nm"));
    EXPECT_EQ(
        "RealLimits.limited(1*deg, 2*deg)",
        Py::Fmt::printRealLimits(RealLimits::limited(1.0 * Units::deg, 2.0 * Units::deg), "rad"));

    // testing printing of real limits in the context of ParameterDistribution and similar
    EXPECT_EQ(", ba.RealLimits.limited(1*deg, 2*deg)",
              Py::Fmt::printRealLimitsArg(RealLimits::limited(1.0 * Units::deg, 2.0 * Units::deg),
                                          "rad"));

    EXPECT_EQ("", Py::Fmt::printRealLimitsArg(RealLimits::limitless()));
}

TEST_F(PythonFormattingTest, printDistribution)
{
    EXPECT_EQ(DistributionGate(1.0, 2.0).pythonConstructor(""), "DistributionGate(1.0, 2.0)");

    EXPECT_EQ(DistributionLorentz(1.0, 2.0).pythonConstructor(""), "DistributionLorentz(1.0, 2.0)");

    EXPECT_EQ(DistributionGaussian(1.0, 2.0).pythonConstructor(""),
              "DistributionGaussian(1.0, 2.0)");

    EXPECT_EQ(DistributionGate(1.0, 2.0).pythonConstructor("nm"), "DistributionGate(1*nm, 2*nm)");

    EXPECT_EQ(DistributionGate(1.0 * Units::deg, 2.0 * Units::deg).pythonConstructor("rad"),
              "DistributionGate(1*deg, 2*deg)");

    EXPECT_EQ(DistributionLogNormal(1.0 * Units::deg, 0.01).pythonConstructor("rad"),
              "DistributionLogNormal(1*deg, 0.01)");

    EXPECT_EQ(DistributionCosine(1.0, 2.0).pythonConstructor(""), "DistributionCosine(1.0, 2.0)");

    EXPECT_EQ(DistributionTrapezoid(1.0, 2.0, 3.0, 4.0).pythonConstructor(""),
              "DistributionTrapezoid(1.0, 2.0, 3.0, 4.0)");
}

TEST_F(PythonFormattingTest, printProfile1D)
{
    EXPECT_EQ(Profile1DCauchy(2.2).pythonConstructor(), "Profile1DCauchy(2.2*nm)");
    EXPECT_EQ(Profile1DGauss(2.2).pythonConstructor(), "Profile1DGauss(2.2*nm)");
    EXPECT_EQ(Profile1DGate(2.2).pythonConstructor(), "Profile1DGate(2.2*nm)");
    EXPECT_EQ(Profile1DTriangle(2.2).pythonConstructor(), "Profile1DTriangle(2.2*nm)");
    EXPECT_EQ(Profile1DCosine(2.2).pythonConstructor(), "Profile1DCosine(2.2*nm)");
    EXPECT_EQ(Profile1DVoigt(2.2, .3).pythonConstructor(), "Profile1DVoigt(2.2*nm, 0.3)");
}

TEST_F(PythonFormattingTest, printProfile2D)
{
    EXPECT_EQ(Profile2DCauchy(2.2, 3.3, 4.4 * Units::deg).pythonConstructor(),
              "Profile2DCauchy(2.2*nm, 3.3*nm, 4.4*deg)");
    EXPECT_EQ(Profile2DGauss(2.2, 3.3, 4.4 * Units::deg).pythonConstructor(),
              "Profile2DGauss(2.2*nm, 3.3*nm, 4.4*deg)");
    EXPECT_EQ(Profile2DGate(2.2, 3.3, 4.4 * Units::deg).pythonConstructor(),
              "Profile2DGate(2.2*nm, 3.3*nm, 4.4*deg)");
    EXPECT_EQ(Profile2DCone(2.2, 3.3, 4.4 * Units::deg).pythonConstructor(),
              "Profile2DCone(2.2*nm, 3.3*nm, 4.4*deg)");
    EXPECT_EQ(Profile2DVoigt(2.2, 3.3, 4.4 * Units::deg, .5).pythonConstructor(),
              "Profile2DVoigt(2.2*nm, 3.3*nm, 4.4*deg, 0.5)");
}

TEST_F(PythonFormattingTest, printParameterDistribution)
{
    DistributionGate gate(1.0, 2.0);
    ParameterDistribution dist(ParameterDistribution::BeamWavelength, gate, 5, 2.0);

    // No RealLimits defined
    EXPECT_EQ(Py::Fmt2::printParameterDistribution(dist, "distr_1"),
              "ba.ParameterDistribution(ba.ParameterDistribution.BeamWavelength, distr_1, 5, 2.0)");

    // RealLimits defined, units unknown
    ParameterDistribution dist2(ParameterDistribution::BeamWavelength, gate, 5, 2.0,
                                RealLimits::limited(1.0, 2.0));
    EXPECT_EQ(Py::Fmt2::printParameterDistribution(dist2, "distr_1"),
              "ba.ParameterDistribution(ba.ParameterDistribution.BeamWavelength, distr_1, "
              "5, 2.0, ba.RealLimits.limited(1.0, 2.0))");

    // RealLimits defined, units explicitly set
    ParameterDistribution dist3(ParameterDistribution::BeamWavelength, gate, 5, 2.0,
                                RealLimits::limited(1.0, 2.0));
    EXPECT_EQ(Py::Fmt2::printParameterDistribution(dist3, "distr_1", "nm"),
              "ba.ParameterDistribution(ba.ParameterDistribution.BeamWavelength, distr_1, "
              "5, 2.0, ba.RealLimits.limited(1*nm, 2*nm))");

    // RealLimits defined, checking that method guess radians units correctly
    ParameterDistribution dist4(ParameterDistribution::BeamAzimuthalAngle, gate, 5, 2.0,
                                RealLimits::limited(1.0 * Units::deg, 2.0 * Units::deg));
    EXPECT_EQ(Py::Fmt2::printParameterDistribution(dist4, "distr_1", "rad"),
              "ba.ParameterDistribution(ba.ParameterDistribution.BeamAzimuthalAngle, "
              "distr_1, 5, 2.0, ba.RealLimits.limited(1*deg, 2*deg))");
}

TEST_F(PythonFormattingTest, printLayerRoughness)
{
    EXPECT_EQ(LayerRoughness(1.1, 2.2, 3.3).pythonConstructor(),
              "LayerRoughness(1.1, 2.2, 3.3*nm)");
}

TEST_F(PythonFormattingTest, printFormFactors)
{
    EXPECT_EQ(FormFactorPyramid2(12.0, 14.0, 5.0, 55 * Units::deg).pythonConstructor(),
              "FormFactorPyramid2(12*nm, 14*nm, 5*nm, 55*deg)");

    EXPECT_EQ(FormFactorGaussSphere(12.0).pythonConstructor(), "FormFactorGaussSphere(12*nm)");

    EXPECT_EQ(FormFactorHemiEllipsoid(6, 7, 5).pythonConstructor(),
              "FormFactorHemiEllipsoid(6*nm, 7*nm, 5*nm)");

    EXPECT_EQ(FormFactorHollowSphere(6, 7).pythonConstructor(),
              "FormFactorHollowSphere(6*nm, 7*nm)");

    EXPECT_EQ(FormFactorBox(6, 7, 5).pythonConstructor(), "FormFactorBox(6*nm, 7*nm, 5*nm)");

    EXPECT_EQ(FormFactorCone(6, 5, 55 * Units::deg).pythonConstructor(),
              "FormFactorCone(6*nm, 5*nm, 55*deg)");

    EXPECT_EQ(FormFactorPyramid6(6, 5, 55 * Units::deg).pythonConstructor(),
              "FormFactorPyramid6(6*nm, 5*nm, 55*deg)");

    EXPECT_EQ(FormFactorBipyramid4(10, 4, 0.7, 55 * Units::deg).pythonConstructor(),
              "FormFactorBipyramid4(10*nm, 4*nm, 0.7, 55*deg)");

    EXPECT_EQ(FormFactorCylinder(3, 5).pythonConstructor(), "FormFactorCylinder(3*nm, 5*nm)");

    EXPECT_EQ(FormFactorDodecahedron(3).pythonConstructor(), "FormFactorDodecahedron(3*nm)");

    EXPECT_EQ(FormFactorEllipsoidalCylinder(3, 4, 5).pythonConstructor(),
              "FormFactorEllipsoidalCylinder(3*nm, 4*nm, 5*nm)");

    EXPECT_EQ(FormFactorCantellatedCube(10, 2).pythonConstructor(),
              "FormFactorCantellatedCube(10*nm, 2*nm)");

    EXPECT_EQ(FormFactorFullSphere(5).pythonConstructor(), "FormFactorFullSphere(5*nm)");

    EXPECT_EQ(FormFactorFullSpheroid(3, 5).pythonConstructor(),
              "FormFactorFullSpheroid(3*nm, 5*nm)");

    EXPECT_EQ(FormFactorIcosahedron(7).pythonConstructor(), "FormFactorIcosahedron(7*nm)");

    EXPECT_EQ(FormFactorLongBoxGauss(6, 7, 5).pythonConstructor(),
              "FormFactorLongBoxGauss(6*nm, 7*nm, 5*nm)");

    EXPECT_EQ(FormFactorLongBoxLorentz(6, 7, 5).pythonConstructor(),
              "FormFactorLongBoxLorentz(6*nm, 7*nm, 5*nm)");

    EXPECT_EQ(FormFactorPrism3(6, 4).pythonConstructor(), "FormFactorPrism3(6*nm, 4*nm)");

    EXPECT_EQ(FormFactorPrism6(3, 4).pythonConstructor(), "FormFactorPrism6(3*nm, 4*nm)");

    EXPECT_EQ(FormFactorPyramid4(10, 4, 60 * Units::deg).pythonConstructor(),
              "FormFactorPyramid4(10*nm, 4*nm, 60*deg)");

    EXPECT_EQ(FormFactorTruncatedSphere(5, 3, 0).pythonConstructor(),
              "FormFactorTruncatedSphere(5*nm, 3*nm, 0*nm)");

    EXPECT_EQ(FormFactorTruncatedSpheroid(3, 5, 1.5, 0).pythonConstructor(),
              "FormFactorTruncatedSpheroid(3*nm, 5*nm, 1.5, 0*nm)");

    EXPECT_EQ(FormFactorPyramid3(16, 4, 60 * Units::deg).pythonConstructor(),
              "FormFactorPyramid3(16*nm, 4*nm, 60*deg)");

    EXPECT_EQ(FormFactorCosineRippleBox(100, 20, 4).pythonConstructor(),
              "FormFactorCosineRippleBox(100*nm, 20*nm, 4*nm)");

    EXPECT_EQ(FormFactorCosineRippleGauss(100, 20, 4).pythonConstructor(),
              "FormFactorCosineRippleGauss(100*nm, 20*nm, 4*nm)");

    EXPECT_EQ(FormFactorCosineRippleLorentz(100, 20, 4).pythonConstructor(),
              "FormFactorCosineRippleLorentz(100*nm, 20*nm, 4*nm)");

    EXPECT_EQ(FormFactorTruncatedCube(15, 6).pythonConstructor(),
              "FormFactorTruncatedCube(15*nm, 6*nm)");

    EXPECT_EQ(FormFactorSawtoothRippleBox(100, 20, 4, 0.3).pythonConstructor(),
              "FormFactorSawtoothRippleBox(100*nm, 20*nm, 4*nm, 0.3*nm)");

    EXPECT_EQ(FormFactorSawtoothRippleGauss(100, 20, 4, 0.3).pythonConstructor(),
              "FormFactorSawtoothRippleGauss(100*nm, 20*nm, 4*nm, 0.3*nm)");

    EXPECT_EQ(FormFactorSawtoothRippleLorentz(100, 20, 4, 0.3).pythonConstructor(),
              "FormFactorSawtoothRippleLorentz(100*nm, 20*nm, 4*nm, 0.3*nm)");

    EXPECT_EQ(FormFactorSphereGaussianRadius(100, 20).pythonConstructor(),
              "FormFactorSphereGaussianRadius(100*nm, 20*nm)");

    EXPECT_EQ(FormFactorBarGauss(100, 20, 4).pythonConstructor(),
              "FormFactorBarGauss(100*nm, 20*nm, 4*nm)");

    EXPECT_EQ(FormFactorBarLorentz(100, 20, 4).pythonConstructor(),
              "FormFactorBarLorentz(100*nm, 20*nm, 4*nm)");

    EXPECT_EQ(FormFactorSphereLogNormalRadius(100, 20, 10).pythonConstructor(),
              "FormFactorSphereLogNormalRadius(100*nm, 20.0, 10)");
}
