#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/PointwiseAxis.h"
#include "Base/Axis/VariableBinAxis.h"
#include "Base/Const/Units.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Model/Device/PointwiseAxisItem.h"
#include "GUI/Util/Error.h"
#include "GUI/View/FromDomain/FromDomain.h"
#include "Tests/GTestWrapper/google_test.h"

class TestAxesItems : public ::testing::Test {
};

TEST_F(TestAxesItems, gettersAndSettes)
{
    BasicAxisItem item;

    item.setBinCount(42);
    EXPECT_EQ(item.binCount(), 42);

    item.setLowerBound(42.1);
    EXPECT_EQ(item.lowerBound(), 42.1);

    item.setUpperBound(42.2);
    EXPECT_EQ(item.upperBound(), 42.2);

    item.setTitle("abc");
    EXPECT_EQ(item.title(), QString("abc"));
}

TEST_F(TestAxesItems, AmplitudeAxisGettersAndSetters)
{
    AmplitudeAxisItem item;

    item.setLogScale(true);
    EXPECT_TRUE(item.isLogScale());
    item.setLogScale(false);
    EXPECT_FALSE(item.isLogScale());
}

TEST_F(TestAxesItems, transformFromDomain)
{
    BasicAxisItem item;

    // transform domain axis without scale factor
    FixedBinAxis axis0("axis0", 99, -42.0, 42.0);
    GUI::Transform::FromDomain::setAxisItem(&item, axis0);
    EXPECT_EQ(item.binCount(), static_cast<int>(axis0.size()));
    EXPECT_EQ(item.lowerBound(), axis0.lowerBound());
    EXPECT_EQ(item.upperBound(), axis0.upperBound());
    EXPECT_EQ(item.title(), QString::fromStdString(axis0.axisName()));

    // transform domain axis with limits given in radians
    FixedBinAxis axis1("axis1", 42, -42.0 * Units::deg, 42.0 * Units::deg);
    const double scale = 1. / Units::deg;
    GUI::Transform::FromDomain::setAxisItem(&item, axis1, scale);
    EXPECT_EQ(item.binCount(), static_cast<int>(axis1.size()));
    EXPECT_EQ(item.lowerBound(), -42.0);
    EXPECT_EQ(item.upperBound(), 42.0);
    EXPECT_EQ(item.title(), QString::fromStdString(axis1.axisName()));

    // transform from unexpected axis
    std::vector<double> bins = {1.0, 2.0, 3.0};
    VariableBinAxis varaxis("axis", bins.size() - 1, bins);
    EXPECT_THROW(GUI::Transform::FromDomain::setAxisItem(&item, varaxis), Error);
}

TEST_F(TestAxesItems, serializeBinaryData)
{
    PointwiseAxisItem axisItem;
    EXPECT_EQ(axisItem.axis(), nullptr);
    EXPECT_EQ(axisItem.serializeBinaryData().size(), 0);

    // init item with pointwise axis
    std::vector<double> coordinates{0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
    PointwiseAxis axis("nameXY", coordinates);
    axisItem.init(axis, "mm");

    // check whether binary data is not empty
    QByteArray binData1 = axisItem.serializeBinaryData();
    EXPECT_NE(axisItem.serializeBinaryData().size(), 0);

    // create a new item, fill it with the serialized data
    PointwiseAxisItem axisItem2;
    EXPECT_EQ(axisItem2.axis(), nullptr);
    axisItem2.deserializeBinaryData(binData1);

    // check whether coordinates are the same as originally set
    EXPECT_NE(axisItem2.axis(), nullptr);
    auto* pwaxis = dynamic_cast<const PointwiseAxis*>(axisItem2.axis());
    ASSERT_NE(pwaxis, nullptr);

    EXPECT_EQ(axis.binCenters(), pwaxis->binCenters());

    // check whether serializing axisItem2 results in same bin data
    QByteArray binData2 = axisItem2.serializeBinaryData();
    EXPECT_EQ(binData1, binData2);
}
